/*----------------------------------------------------------------------------
**   Copyright (C) 1987 Thomas Taylor
**
** This file is part of TexDraw
**
** TexDraw is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY.  No author or distributor
** accepts responsibility to anyone for the consequences of using it
** or for whether it serves any particular purpose or works at all,
** unless he says so in writing. 
**
** This software is free of charge.  It may not be re-sold.
** Everyone is granted permission to copy, modify and redistribute
** TexDraw, as long as the original copyright and this notice are
** retained.
**
**
**         Module:  texdraw.h
**
**   Program Name:  texdraw
**    Program Use:  Graphics Program for LaTeX Text Formatter
**
**         Author:  Tom Taylor
**           Date:  August 21, 1987
**
**          Class:  C Sci 498, Senior Project
**        Sponsor:  Hank Levy
**
**
**    This module contains headers and definitions for the texdraw program.
**    The program code is contained in the following modules:
**
**    texdriv.c     the driver
**    texgeom.c     LaTeX geometry functions
**    texobj.c      individual object functions
**    texobjs.c     object set functions
**    texpoly.c     polygon functions
**    texread.c     LaTeX string parsing functions
**    texstr.c      string functions
**    texwind.c     display and window functions.
**
----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------
**   Copyright (C) 1988 Hiroto Kagotani
----------------------------------------------------------------------------*/


#include <math.h>
#include <stdio.h>
#include <ctype.h>
#include <strings.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/X10.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>

#define ABS(A)         ((A) < 0 ? -(A) : (A))
#define MAX(A,B)       ((A) > (B) ? (A) : (B))
#define MIN(A,B)       ((A) < (B) ? (A) : (B))
#define ODD(A)         ((A) & 01)
#define EVEN(A)        (!ODD(A))
#define SIGN(A)        ( (A) < 0 ? -1 : 1 )

#define ROUND(A)       ( (A) < 0 ? (int)((A) - 0.5) : (int)((A) + 0.5) )

#define ROUND1(A)      ( (A) < 0 ? (((int)(10 * (A) - 0.5)) / 10.0) : \
                                   (((int)(10 * (A) + 0.5)) / 10.0) )

#define ROUND2(A)      ( (A) < 0 ? (((int)(100 * (A) - 0.5)) / 100.0) : \
                                   (((int)(100 * (A) + 0.5)) / 100.0) )

#define LENGTH(dx,dy)  (sqrt((double)(dx)*(dx)+(dy)*(dy)))

#define SWAP(a,b)   {int TMP; TMP = (a); (a) = (b); (b) = TMP;}
#define SORT2(a,b)  if ((a) > (b)) SWAP((a),(b))

#define CLOSER(A,B,X)          (ABS((A)-(X)) < ABS((B)-(X)) ? (A) : (B))

#define CLOSEST(A,B,C,X)       (ABS((A)-(X)) < ABS((B)-(X)) ? \
                                ABS((A)-(X)) < ABS((C)-(X)) ? (A) : (C) : \
                                ABS((B)-(X)) < ABS((C)-(X)) ? (B) : (C))

#define MAXINT     0x7fffffff       /*  2,147,836,647 */
#define MININT     0x80000000       /* -2,147,836,648 */

#define INV(Y)	(picHeight-1 - (Y))

/* --------------------------------------------------------------------------
   Define the boolean type and constants.
---------------------------------------------------------------------------*/

typedef short bool;

#define FALSE      0
#define TRUE       1

#define NONE      -1
#define VALID(X)       ((X) != NONE)

#define STATIC     0
#define CURRENT    1
#define DELETED    2

#define LINEBUFSIZE 132
#define CMDBUFSIZE  132
#define STRBUFSIZE  132

#define CTRL_C      3        /* control-C       */
#define BS          8        /* backspace       */
#define LF         10        /* line feed       */
#define CR         13        /* carriage return */
#define SP         32        /* space           */
#define DEL       127        /* delete          */

/* default spacing for grid lines */
#define DEFAULTGRID 20

/* --------------------------------------------------------------------------
   Define object codes and menu command numbers.
   The order of assignment here determines where the commands appear
   in the array of menu windows.
---------------------------------------------------------------------------*/

/* primary modes */
#define LINE            0
#define BOX             1
#define CIRCLE          2

#define VECTOR          3
#define RULE            4
#define DISK            5

#define OVAL            7
#define TEXT            8

#define MOVEORIGIN      9
#define MOVEPOINT      10
#define RESIZE         11

#define SELECTCURRENT  12
#define MOVECURRENT    13
#define COPYCURRENT    14

/* secondary modes */
#define GRAVITY        18
#define ALIGNH         19
#define ALIGNV         20

#define THICK          21
#define FONTS          22

#define GRID           24
#define GRIDSIZE       25
#define SNAPGRID       26

/* commands */
#define READ           30
#define WRITE          31
#define DUMPDATA       32

#define SHOWCURRENT    33
#define REDRAW         34
#define UNDO           35

#define DELETECURRENT  36
#define CLEAR          37

#define QUIT           41


/* --------------------------------------------------------------------------
   Define the kinds of objects.
---------------------------------------------------------------------------*/

typedef struct _Line {
  short dx, dy;                                   /* length components     */
  short tx, ty;                                   /* TEX slope definition  */
  short len;                                      /* TEX length definition */
  short thick;                                    /* line thickness        */
} Line;


typedef struct _Vector {
  short dx, dy;                                   /* length components     */
  short tx, ty;                                   /* TEX slope definition  */
  short len;                                      /* TEX length definition */
  short thick;                                    /* line thickness        */
} Vector;


typedef struct _Box {
  short w, h;                                     /* width, height         */
  short thick;                                    /* line thickness        */
} Box;


typedef struct _Rule {
  short w, h;                                     /* width, height         */
} Rule;


#define OVAL_noT   1
#define OVAL_noL   2
#define OVAL_noB   4
#define OVAL_noR   8
typedef struct _Oval {
  short w, h;                                     /* width, height         */
  short part;                                     /* part of oval to draw  */
  short thick;                                    /* line thickness        */
} Oval;


typedef struct _Circle {
  int   dia;                                      /* diameter of circle    */
  short thick;                                    /* line thickness        */
} Circle;


typedef struct _Disk {
  int   dia;                                      /* diameter of disk      */
} Disk;


typedef struct _Text {
  char     *txtptr;                               /* string contents       */
  short     txtlen;                               /* length of string      */
  short     w, h;                                 /* width, height         */
  XFontStruct *font;                              /* X font info pointer   */
  short     fontNum;                              /* index into font list  */
} Text;




/* --------------------------------------------------------------------------
   Define the type "Object".
---------------------------------------------------------------------------*/

#define STATIC     0
#define CURRENT    1
#define DELETED    2

typedef struct _Object {
  short code;                   /* what kind of object is it?              */
  short status;                 /* STATIC, CURRENT, DELETED                */
  short x, y;                   /* position of the object:                 */
                                /*   line: start;  box, text: lower left   */
                                /*   oval, circle, disk: center            */
  union {
    Line   line;
    Vector vector;
    Box    box;
    Rule   rule;
    Oval   oval;
    Circle circle;
    Disk   disk;
    Text   text;
  } kind;
} Object;




bool driv_EnterCommand();
void driv_EnterPicText();
void driv_ExecuteCommand();
bool driv_IsLeftButton();
void driv_MenuSelectCmd();
void driv_RepeatCommand();
void driv_RubberBand();
void driv_Quit();


void gc_MakeFilled();
void gc_MakeFilledErase();
void gc_MakeGrid();
void gc_MakeSelect();
GC   gc_MakeTextErase();
GC   gc_MakeTextWrite();
void gc_MakeThick();
void gc_MakeThickErase();
void gc_MakeThin();
void gc_MakeThinErase();
void gc_MakeTiled();
void gc_MakeXorDashed();
void gc_MakeXorDotted();


void geom_BestTexObject();
void geom_GraphToTexRad();
void geom_GraphToTexSlope();
void geom_TexToGraphSlope();

int  obj_ClosestX();
int  obj_ClosestY();
void obj_Copy();
void obj_CreateGeom();
void obj_CreateText();
void obj_Delete();
void obj_Destroy();
void obj_DrawObj();
void obj_DumpObj();
void obj_EraseObj();
XFontStruct *obj_FontOf();
void obj_GetCoordsOfPoint();
bool obj_Gravitate();
bool obj_GravitateOpposite();
bool obj_IsBounded();
bool obj_IsCurrent();
bool obj_IsDeleted();
bool obj_IsNearObj();
bool obj_IsNearPt();
int  obj_NearPointNo();
int  obj_OppositePointNo();
void obj_Move();
void obj_MovePoint();
void obj_ReadFromFile();
void obj_SetCurrent();
void obj_SetGridTile();
void obj_SetThickness();
int  obj_ShapeOf();
int  obj_ThicknessOf();
void obj_ToggleGravity();
void obj_ToggleThickness();
void obj_WriteToFile();
void obj_UnsetCurrent();

void objs_AddGeom();
void objs_AddText();
void objs_Align();
bool objs_AreBoundedBy();
void objs_ConditionalRedraw();
void objs_DeleteCurrent();
void objs_DeleteAll();
void objs_DrawCurrent();
void objs_DrawAll();
void objs_DumpSet();
void objs_DuplicateCurrent();
void objs_EraseDeletions();
void objs_Gravitate();
void objs_GravitateOpposite();
void objs_InitShape();
void objs_MoveCurrent();
void objs_MovePoint();
void objs_PurgeDeletions();
void objs_ReadFromFile();
void objs_Redraw();
void objs_SetCurrentIfBounded();
void objs_ShowCurrent();
void objs_SwapCurrentWithDeletions();
void objs_ToggleHoriz();
void objs_ToggleVert();
void objs_Undo();
void objs_UnsetCurrent();
void objs_Update();
void objs_WriteToFile();

void poly_Dump();
void poly_Rubber();
void poly_BoxOnGraph();
void poly_Box();
void poly_Circle();
void poly_Disk();
void poly_Line();
void poly_Oval();
void poly_Vector();

void read_GetCoords();
void read_GetNum();
void read_GetObject();
void read_GetOvalOption();
void read_GetPicture(); 
void read_GetReal(); 
void read_GetRealCoords(); 
void read_GetString();
void read_GetSym();
bool read_IsCommandChar();
bool read_IsDelimiter();
bool read_IsDigit();
bool read_IsSpace();
bool read_IsStringChar();
void read_MatchSym();
bool read_ParseLine();
void read_PrintSym();

int  str_Length();
void str_Copy();
bool str_IsEqual();

void wind_ChangeFont();
void wind_DefinePixels();
void wind_ErrorMsg1();
void wind_ErrorMsg2();
void wind_Finalize();
void wind_GetScreenOffsets();
void wind_GetTexOffsets();
void wind_HandleExposure();
void wind_Initialize();
XFontStruct *wind_LookupFontFromTex();
int  wind_LookupFontNum();
int  wind_LookupMenuNum();
Window wind_LookupMenuWindow();
char *wind_LookupTexFromFont();
GC   wind_LookupTextGC();
GC   wind_LookupTextEraseGC();
void wind_MakeGraphics();
void wind_MakeMenus();
void wind_MakePicture();
void wind_MakeSmallCrossCursor();
void wind_MakeText();
void wind_Message1();
void wind_Message2();
void wind_OpenDisplay();
void wind_OpenFonts();
void wind_PaintMenu();
void wind_PaintMenuNum();
void wind_PaintMenuWind();
void wind_PrintXFontStruct();
void wind_PrintWindow();
void wind_ReportPictureSize();
void wind_ResizePicture();
void wind_SetPicFont();
void wind_SelectMenu();
void wind_UnselectMenu();

#define rootWind	RootWindow(display, screen)

extern char *programName, *displayName;
extern Display *display;
extern Window picWind, textWind, graphWind;
extern int screen;
extern Cursor smallCrossCursor;
extern XFontStruct *picFont, *textFont;
extern GC selectGC, tiledGC,
	  gridGC, thickGC, thickEraseGC, thinGC, thinEraseGC,
	  filledGC, filledEraseGC, xorDottedGC, xorDashedGC;
extern Pixmap gridTile;

extern unsigned long fore_pixel, back_pixel, brdr_pixel;

extern int picWidth, picHeight;
extern int thickness;
extern double scale;

extern int scrXofs, scrYofs;
extern double texXofs, texYofs;

extern int currFontNum;
extern int gridSize;
