/*----------------------------------------------------------------------------
**   Copyright (C) 1987 Thomas Taylor
**
** This file is part of TexDraw
**
** TexDraw is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY.  No author or distributor
** accepts responsibility to anyone for the consequences of using it
** or for whether it serves any particular purpose or works at all,
** unless he says so in writing. 
**
** This software is free of charge.  It may not be re-sold.
** Everyone is granted permission to copy, modify and redistribute
** TexDraw, as long as the original copyright and this notice are
** retained.
**
**
**         Module:  texgeom.c
**
**   Program Name:  texdraw
**    Program Use:  Graphics Program for LaTeX Text Formatter
**
**         Author:  Tom Taylor
**           Date:  August 21, 1987
**
**          Class:  C Sci 498, Senior Project
**        Sponsor:  Hank Levy
**
**
**    This module contains LaTeX geometry functions for the texdraw program.
**
----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------
**   Copyright (C) 1988 Hiroto Kagotani
----------------------------------------------------------------------------*/


#include "texdraw.h"


#define MAXDIM 2000


/*----------------------------------------------------------------------------
  The array "lineDefs" contains integers between 0 and 6 which are used in
  pairs to define allowable LaTeX line slopes in the first quadrant.
  The integers are ordered such that:
    if n = sizeof(lineDefs) - 1,
    then for all i, 0 <= i < n,
    tx[i] = lineDefs[i] and ty[i] = lineDefs[n-i]
    and (ty[i] / tx[i]) < (ty[i+1] / tx[i+1])
----------------------------------------------------------------------------*/
static int lineDefs[] = {
  1, 6, 5, 4, 3, 5, 2, 5, 3, 4, 5, 6, 1, 5, 4, 3, 2, 3, 1, 2, 1, 1, 1, 1, 0
};



/*----------------------------------------------------------------------------
  The array "vectorDefs" contains integers between 0 and 4 which are used in
  pairs to define allowable LaTeX vector slopes in the first quadrant.
  The integers are ordered such that:
    if n = sizeof(vectorDefs) - 1,
    then for all i, 0 <= i < n,
    tx[i] = vectorDefs[i] and ty[i] = vectorDefs[n-i]
    and (ty[i] / tx[i]) < (ty[i+1] / tx[i+1])
----------------------------------------------------------------------------*/
static int vectorDefs[] = {
  1, 4, 3, 2, 3, 4, 1, 3, 2, 1, 1, 1, 0
};



/*----------------------------------------------------------------------------
  The array "lineSlopes" contains slopes midway between allowable LaTeX
  line slopes.
----------------------------------------------------------------------------*/
static double lineSlopes[] = {
   .0833,  .1833,  .2250,  .2917,  .3667,   .4500,
   .5500,  .6333,  .7083,  .7750,  .8167,   .9167,
  1.1000, 1.2250, 1.2917, 1.4167, 1.5833,  1.8333,
  2.2500, 2.7500, 3.5000, 4.5000, 5.5000, 12.0000,
  MAXINT
};



/*----------------------------------------------------------------------------
  The array "vectorSlopes" contains slopes 'midway' between allowable LaTeX
  vector slopes.
----------------------------------------------------------------------------*/
static double vectorSlopes[] = {
   .1250,  .2917,  .4167,  .5834,
   .7084,  .8750, 1.1667, 1.4167,
  1.7500, 2.5000, 3.5000, 8.0000,
  MAXINT
};



/*----------------------------------------------------------------------------
  The array "circleRadii" contains the radii for allowable LaTeX circles
  with even diameters.
----------------------------------------------------------------------------*/
static int circleRadii[] = {
  1, 2, 3, 4, 5, 6, 7, 8, 10, 12, 14, 16, 18, 20, MAXINT
};



/*----------------------------------------------------------------------------
  The array "diskRadii" contains the radii for allowable LaTeX disks
  with even diameters.
----------------------------------------------------------------------------*/
static int diskRadii[] = {
  1, 2, 3, 4, 5, 6, 7, MAXINT
};



static double *slopes;                 /* line or vector slopes array       */
static int *defs;                      /* line or vector definitions array  */
static int *radii;                     /* circle or disk radii array        */

static int maxSlopes;                  /* highest index into slopes[]       */
static int maxDefs;                    /* highest index into defs[]         */
static int maxRadii;                   /* highest index into radii[]        */




/*----------------------------------------------------------------------------
  geom_BestTexObject:
    If "cmd" identifies an object with a limited selection of allowable
    LaTeX representations, then x2 and y2 are recalculated to values for
    the nearest allowable object.
----------------------------------------------------------------------------*/
void
geom_BestTexObject(cmd, x1, y1, x2, y2)
int cmd, x1, y1, *x2, *y2;
{
  int dx, dy, tx, ty, length, rad;

  *x2 = MAX(*x2, 0);   *x2 = MIN(*x2, MAXDIM);
  *y2 = MAX(*y2, 0);   *y2 = MIN(*y2, MAXDIM);

  switch (cmd) {
    case CIRCLE:
    case DISK:
      rad = ROUND(LENGTH(*x2-x1, *y2-y1));
      geom_GraphToTexRad(cmd, &rad);
      *x2 = x1 + rad;
      *y2 = y1;
      break;
    case LINE:
    case VECTOR:
      dx = *x2 - x1;
      dy = *y2 - y1;
      geom_GraphToTexSlope(cmd, &dx, &dy, &tx, &ty, &length);
      *x2 = x1 + dx;
      *y2 = y1 + dy;
      break;
  }
}




/*----------------------------------------------------------------------------
  geom_GraphToTexRad:
    Recomputes "rad" to the nearest allowable value for a LaTeX circle
    or disk.
----------------------------------------------------------------------------*/
void
geom_GraphToTexRad(cmd, rad)
int cmd, *rad;
{
  int i = 0;
  double fRad = scale * *rad;                              /* scaled radius */

  /* select which array to use, and get its size */
  switch (cmd) {
    case CIRCLE:
      radii = circleRadii;
      maxRadii = sizeof(circleRadii) / sizeof(int) - 1;
      break;
    case DISK:
      radii = diskRadii;
      maxRadii = sizeof(diskRadii) / sizeof(int) - 1;
      break;
  }

  /* search for the nearest LaTeX radius */
  fRad = MIN(fRad, (double) MAXINT);
  while (fRad >= (double) radii[i+1]) i++;

  /* Calculate the unscaled screen radius. */
  *rad = ROUND(radii[i] / scale);
}




/*----------------------------------------------------------------------------
  geom_GraphToTexSlope:
    Given graphics parameters dx and dy:
    1.  Computes nearest TeX parameters tx, ty and len.
    2.  Recomputes dx and dy from tx, ty, and len.
----------------------------------------------------------------------------*/
void
geom_GraphToTexSlope(cmd, dx, dy, tx, ty, length)
int cmd, *dx, *dy, *tx, *ty, *length;
{
  int i = 0;
  int xSign, ySign;
  double slope;

  /* convert to first quadrant (work only with positive numbers) */
  xSign = SIGN(*dx);
  ySign = SIGN(*dy);
  *dx = ABS(*dx);
  *dy = ABS(*dy);

  /* select which arrays to use, and get their sizes */
  switch (cmd) {
    case LINE:
      slopes = lineSlopes;
      defs = lineDefs;
      maxSlopes = sizeof(lineSlopes) / sizeof(double) - 1;
      maxDefs = sizeof(lineDefs) / sizeof(int) - 1;
      break;
    case VECTOR:
      slopes = vectorSlopes;
      defs = vectorDefs;
      maxSlopes = sizeof(vectorSlopes) / sizeof(double) - 1;
      maxDefs = sizeof(vectorDefs) / sizeof(int) - 1;
      break;
  }

  /* find the graphics slope */
  slope = (*dx == 0) ? (double) MAXINT : *dy / (double) *dx;
  slope = MIN(slope, MAXINT);

  /* search for the nearest LaTeX slope */
  while (slope > slopes[i]) i++;

  /* assign the LaTeX definitions for the selected slope */
  /* and restore signs for the appropriate quadrant      */
  *tx = defs[i];
  *ty = defs[maxDefs-i];

  if      (*tx == 0)   *dx = 0;
  else if (*ty == 0)   *dy = 0;
  else if (*dx > *dy)  *dy = (*dx * *ty) / *tx;
  else                 *dx = (*dy * *tx) / *ty;

  *length = (*tx == 0) ? *dy : *dx;
  *tx *= xSign;
  *ty *= ySign;

  /* recompute dx and dy from tx, ty, and length */
  geom_TexToGraphSlope(*tx, *ty, *length, dx, dy);

/*****
  *length = (*length < 10) ? 0 : *length;
*****/

}




/*----------------------------------------------------------------------------
  geom_TexToGraphSlope:
    Given TeX parameters tx, ty, and len:
      Computes graphics parameters dx and dy.
----------------------------------------------------------------------------*/
void
geom_TexToGraphSlope(tx, ty, len, dx, dy)
int tx, ty, len, *dx, *dy;
{
  if (tx == 0) {
    *dx = 0;
    *dy = SIGN(ty) * len;
  }
  else {
    *dx = SIGN(tx) * len;
    *dy = ROUND(len * ty / (double) ABS(tx));
  }
}


