/*----------------------------------------------------------------------------
**   Copyright (C) 1987 Thomas Taylor
**
** This file is part of TexDraw
**
** TexDraw is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY.  No author or distributor
** accepts responsibility to anyone for the consequences of using it
** or for whether it serves any particular purpose or works at all,
** unless he says so in writing. 
**
** This software is free of charge.  It may not be re-sold.
** Everyone is granted permission to copy, modify and redistribute
** TexDraw, as long as the original copyright and this notice are
** retained.
**
**
**         Module:  texobj.c
**
**   Program Name:  texdraw
**    Program Use:  Graphics Program for LaTeX Text Formatter
**
**         Author:  Tom Taylor
**           Date:  August 21, 1987
**
**          Class:  C Sci 498, Senior Project
**        Sponsor:  Hank Levy
**
**
**    This module contains individual object functions for the texdraw
**    program.
**
----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------
**   Copyright (C) 1988 Hiroto Kagotani
----------------------------------------------------------------------------*/

#include "texdraw.h"

int thickness = 0;

static int gravity = 0;

int grid = 0;
int snapToGrid = 0;
int gridSize = DEFAULTGRID;
Pixmap gridTile = None;

extern double unitLength;

extern char *malloc();
#define alloc(n,type)  ((type *) malloc((unsigned)(n) * sizeof(type)))
#define dispose(p)     {free((char *) p); p=0; }


/* --------------------------------------------------------------------------
   obj_ClosestX:
     Returns the x-coordinate (if any) the object point which is
     horizontally closest to "x".
---------------------------------------------------------------------------*/
int
obj_ClosestX(x, obj)
int x;
Object obj;
{
  int c = obj.code;
  int a = obj.x;
  int b, r;

  switch (c) {
    case BOX:       b = a + obj.kind.box.w;         break;
    case CIRCLE:    r = obj.kind.circle.dia / 2;    break;
    case DISK:      r = obj.kind.disk.dia / 2;      break;
    case LINE:      b = a + obj.kind.line.dx;       break;
    case OVAL:      b = a + obj.kind.oval.w;        break;
    case RULE:      b = a + obj.kind.rule.w;        break;
    case TEXT:      b = a + obj.kind.text.w;        break;
    case VECTOR:    b = a + obj.kind.vector.dx;     break;
  }

  if (c==BOX || c==LINE || c==OVAL || c==RULE || c==TEXT || c==VECTOR)
    return CLOSER(a, b, x);

  else if (c==CIRCLE || c==DISK)
    return CLOSEST(a, a+r, a-r, x);

  else
    return NONE;
}




/* --------------------------------------------------------------------------
   obj_ClosestY:
     Returns the y-coordinate (if any) the object point which is
     vertically closest to "y".
---------------------------------------------------------------------------*/
int
obj_ClosestY(y, obj)
int y;
Object obj;
{
  int c = obj.code;
  int a = obj.y;
  int b, r;

  switch (c) {
    case BOX:       b = a + obj.kind.box.h;         break;
    case CIRCLE:    r = obj.kind.circle.dia / 2;    break;
    case DISK:      r = obj.kind.disk.dia / 2;      break;
    case LINE:      b = a + obj.kind.line.dy;       break;
    case OVAL:      b = a + obj.kind.oval.h;        break;
    case RULE:      b = a + obj.kind.rule.h;        break;
    case TEXT:      b = a + obj.kind.text.h;        break;
    case VECTOR:    b = a + obj.kind.vector.dy;     break;
  }

  if (c==BOX || c==LINE || c==OVAL || c==RULE || c==TEXT || c==VECTOR)
    return CLOSER(a, b, y);

  else if (c==CIRCLE || c==DISK)
    return CLOSEST(a, a+r, a-r, y);

  else
    return NONE;
}




/* --------------------------------------------------------------------------
  obj_Copy:
    Copy object "a" to object "b".
    If "a" is a TEXT object, allocate memory for b's string, and copy a's
    string into it.
---------------------------------------------------------------------------*/
void
obj_Copy(a, b)
Object a, *b;
{
  *b = a;
  if (a.code == TEXT) {
    unsigned n;

    n = 1 + a.kind.text.txtlen;
    b->kind.text.txtptr = alloc(n, char);
    str_Copy(a.kind.text.txtptr, b->kind.text.txtptr);
  }
}




/* --------------------------------------------------------------------------
   obj_CreateGeom:
     Fill in the structure for a geometric object.
---------------------------------------------------------------------------*/

void
obj_CreateGeom(cmd, x1, y1, x2, y2, obj)
int cmd, x1, y1, x2, y2;
Object *obj;
{
  int rad = ABS(x2 - x1);
  int dx = x2 - x1;
  int dy = y2 - y1;
  int tx, ty, len;

  obj->status = CURRENT;
  switch (cmd) {
    case BOX:
      obj->code = BOX;
      obj->x = (short)MIN(x1, x2);
      obj->y = (short)MIN(y1, y2);
      obj->kind.box.w = (short)ABS(dx);
      obj->kind.box.h = (short)ABS(dy);
      obj->kind.box.thick = (short)thickness;
      break;
    case CIRCLE:
      geom_GraphToTexRad(cmd, &rad);
      obj->code = CIRCLE;
      obj->x = (short)x1;
      obj->y = (short)y1;
      obj->kind.circle.dia = (short) 2 * rad;
      obj->kind.circle.thick = (short)thickness;
      break;
    case DISK:
      geom_GraphToTexRad(cmd, &rad);
      obj->code = DISK;
      obj->x = (short)x1;
      obj->y = (short)y1;
      obj->kind.disk.dia = (short) 2 * rad;
      break;
    case LINE:
      geom_GraphToTexSlope(cmd, &dx, &dy, &tx, &ty, &len);
      obj->code = LINE;
      obj->x = (short)x1;
      obj->y = (short)y1;
      obj->kind.line.dx = (short)dx;
      obj->kind.line.dy = (short)dy;
      obj->kind.line.tx = (short)tx;
      obj->kind.line.ty = (short)ty;
      obj->kind.line.len = (short)len;
      obj->kind.line.thick = (short)thickness;
      break;
    case MOVECURRENT:
      obj->code = LINE;
      obj->x = (short)x1;
      obj->y = (short)y1;
      obj->kind.line.dx = (short)dx;
      obj->kind.line.dy = (short)dy;
      obj->kind.line.tx = (short)tx;
      obj->kind.line.ty = (short)ty;
      obj->kind.line.len = (short)len;
      obj->kind.line.thick = (short)thickness;
      break;
    case OVAL:
      obj->code = OVAL;
      obj->x = (short)MIN(x1, x2);
      obj->y = (short)MIN(y1, y2);
      obj->kind.oval.w = (short)ABS(dx);
      obj->kind.oval.h = (short)ABS(dy);
      obj->kind.oval.thick = (short)thickness;
      break;
    case RULE:
      obj->code = RULE;
      obj->x = (short)MIN(x1, x2);
      obj->y = (short)MIN(y1, y2);
      obj->kind.rule.w = (short)ABS(dx);
      obj->kind.rule.h = (short)ABS(dy);
      break;
    case VECTOR:
      geom_GraphToTexSlope(cmd, &dx, &dy, &tx, &ty, &len);
      obj->code = VECTOR;
      obj->x = (short)x1;
      obj->y = (short)y1;
      obj->kind.vector.dx = (short)dx;
      obj->kind.vector.dy = (short)dy;
      obj->kind.vector.tx = (short)tx;
      obj->kind.vector.ty = (short)ty;
      obj->kind.vector.len = (short)len;
      obj->kind.vector.thick = (short)thickness;
      break;
  }
}




/* --------------------------------------------------------------------------
   obj_CreateText:
     1.  Allocate memory for the text string.
     2.  Fill in the structure for a text object.
---------------------------------------------------------------------------*/

void
obj_CreateText(x, y, str, len, font, obj)
int x, y;
char *str;
int len;
XFontStruct *font;
Object *obj;
{
  unsigned n = len;
  int w, h;
  int fontNum = wind_LookupFontNum(font);

  if (!VALID(fontNum))
    wind_ErrorMsg1("CreateText:  no font");

  /* null-terminate the string, and calculate its size in pixels */
  str[len] = '\0';
  h = font->ascent;
  w = XTextWidth(font, str, len);

  obj->code = TEXT;
  obj->status = CURRENT;
  obj->x = (short)x;
  obj->y = (short)y;
  obj->kind.text.w = (short)w;
  obj->kind.text.h = (short)h;
  obj->kind.text.txtlen = (short)n;
  obj->kind.text.txtptr = alloc(n+1, char);
  obj->kind.text.font = font;
  obj->kind.text.fontNum = fontNum;
  str_Copy(str, obj->kind.text.txtptr);
}




/* --------------------------------------------------------------------------
   obj_Delete:
     Delete an object.
---------------------------------------------------------------------------*/

void
obj_Delete(obj)
Object *obj;
{
  obj->status = DELETED;
}




/* --------------------------------------------------------------------------
   obj_Destroy:
     If an object is a TEXT object, free memory allocated for its string.
---------------------------------------------------------------------------*/

void
obj_Destroy(obj)
Object *obj;
{
  if (obj->code == TEXT)
    free(obj->kind.text.txtptr);
}




/* --------------------------------------------------------------------------
   obj_DrawObj:
     Draw an object in the picture window.
---------------------------------------------------------------------------*/

void
obj_DrawObj(obj)
Object obj;
{
  int x1, y1, x2, y2, xofs, yofs;
  int x, y, n;
  char *t;
  Text text;
  XFontStruct *font;

  switch (obj.code) {
    case BOX:
      poly_Box(obj, obj.kind.box.thick ? thickGC : thinGC);
      break;

    case RULE:
      poly_Rule(obj, tiledGC);
      break;

    case LINE:
      poly_Line(obj, obj.kind.line.thick ? thickGC : thinGC);
      break;

    case CIRCLE:
      poly_Circle(obj, obj.kind.circle.thick ? thickGC : thinGC);
      break;

    case DISK:
      poly_Disk(obj, tiledGC);
      break;

    case OVAL:
      poly_Oval(obj, obj.kind.oval.thick ? thickGC : thinGC);
      break;

    case VECTOR:
      poly_Vector(obj, obj.kind.vector.thick ? thickGC : thinGC);
      break;

    case TEXT:
      text = obj.kind.text;
      font = text.font;
      wind_GetScreenOffsets(text.fontNum, &xofs, &yofs);
      x = obj.x + xofs;
      y = obj.y + yofs;
      t = text.txtptr;
      n = text.txtlen;
      XDrawString(display, picWind, wind_LookupTextGC(font), x, INV(y), t, n);
      break;
  }
}




#ifdef DUMPDATA
/* --------------------------------------------------------------------------
   Dump the contents of an object (for debug purposes).
---------------------------------------------------------------------------*/

void
obj_DumpObj(obj)
Object obj;
{
  switch (obj.code) {
    case    BOX: printf("BOX    \n"); break;
    case CIRCLE: printf("CIRCLE \n"); break;
    case   DISK: printf("DISK   \n"); break;
    case   LINE: printf("LINE   \n"); break;
    case   OVAL: printf("OVAL   \n"); break;
    case   TEXT: printf("TEXT   \n"); break;
    case VECTOR: printf("VECTOR \n"); break;
  }

  printf("    code: %4d      \n", obj.code     );

  printf("  status: ");
  switch (obj.status) {
    case STATIC:  printf("STATIC  \n"); break;
    case CURRENT: printf("CURRENT \n"); break;
    case DELETED: printf("DELETED \n"); break;
  }

  printf("   x,  y: %4d, %4d \n", obj.x, obj.y );

  switch (obj.code) {
    case BOX: {
      Box b;
      b = obj.kind.box;
      printf("   w,  h: %4d, %4d \n", b.w, b.h);
      printf("   thick: %4d \n", b.thick);
      break;
    }
    case CIRCLE: {
      Circle c;
      c = obj.kind.circle;
      printf("     dia: %4d \n", c.dia);
      printf("   thick: %4d \n", c.thick);
      break;
    }
    case DISK: {
      Disk d;
      d = obj.kind.disk;
      printf("     dia: %4d \n", d.dia);
      break;
    }
    case LINE: {
      Line l;
      l = obj.kind.line;
      printf("  dx, dy: %4d, %4d \n", l.dx, l.dy );
      printf("  tx, ty: %4d, %4d \n", l.tx, l.ty );
      printf("     len: %4d \n", l.len);
      printf("   thick: %4d \n", l.thick);
      break;
    }
    case OVAL: {
      Oval o;
      o = obj.kind.oval;
      printf("   w,  h: %4d, %4d \n", o.w, o.h);
      printf("    part: %s  \n", o.part);
      printf("   thick: %4d \n", o.thick);
      break;
    }
    case RULE: {
      Rule r;
      r = obj.kind.rule;
      printf("   w,  h: %4d, %4d \n", r.w, r.h);
      break;
    }
    case TEXT: {
      Text t;
      t = obj.kind.text;
      printf("   w,  h: %4d, %4d \n", t.w, t.h);
      printf("    text: %s  \n", t.txtptr);
      printf("  txtlen: %4d \n", t.txtlen);
      printf(" fontNum: %4d \n", t.fontNum);
      printf("    font: %s  \n", wind_LookupTexFromFont(t.font));
      break;
    }
    case VECTOR: {
      Vector v;
      v = obj.kind.vector;
      printf("  dx, dy: %4d, %4d \n", v.dx, v.dy );
      printf("  tx, ty: %4d, %4d \n", v.tx, v.ty );
      printf("     len: %4d \n", v.len);
      printf("   thick: %4d \n", v.thick);
      break;
    }
  }
}
#endif




/* --------------------------------------------------------------------------
   obj_EraseObj:
     Erase an object in the picture window.
---------------------------------------------------------------------------*/

void
obj_EraseObj(obj)
Object obj;
{
  int x1, y1, x2, y2, thick, xofs, yofs;
  Line line;
  Box box;
  Rule rule;
  Oval oval;
  Circle circle;
  Disk disk;
  Vector vector;
  int x, y, n;
  char *t;
  Text text;
  XFontStruct *font;

  switch (obj.code) {
    case BOX:
      poly_Box(obj, box.thick ? thickEraseGC : thinEraseGC);
      break;

    case RULE:
      poly_Rule(obj, filledEraseGC);
      break;

    case LINE:
      poly_Line(obj, line.thick ? thickEraseGC : thinEraseGC);
      break;

    case CIRCLE:
      poly_Circle(obj, circle.thick ? thickEraseGC : thinEraseGC);
      break;

    case DISK:
      poly_Disk(obj, filledEraseGC);
      break;

    case OVAL:
      poly_Oval(obj, oval.thick ? thickEraseGC : thinEraseGC);
      break;

    case VECTOR:
      poly_Vector(obj, vector.thick ? thickEraseGC : thinEraseGC);
      break;

    case TEXT:
      text = obj.kind.text;
      font = text.font;
      wind_GetScreenOffsets(text.fontNum, &xofs, &yofs);
      x = obj.x + xofs;
      y = obj.y + yofs;
      t = text.txtptr;
      n = text.txtlen;
      XDrawString(display, picWind, wind_LookupTextEraseGC(font),
							x, INV(y), t, n);
      break;
  }
}




/* --------------------------------------------------------------------------
   obj_FontOf:
---------------------------------------------------------------------------*/
XFontStruct *
obj_FontOf(obj)
Object obj;
{
  if (obj.code == TEXT)
    return obj.kind.text.font;
  else
    return NULL;
}




/* --------------------------------------------------------------------------
   obj_GetCoordsOfPoint:
     Gets the coordinates (x, y) of point number "n" of an object.
---------------------------------------------------------------------------*/
void
obj_GetCoordsOfPoint(n, x, y, obj)
int n, *x, *y;
Object obj;
{
  *x = obj.x;
  *y = obj.y;
  if (n == 0) return;

  switch (obj.code) {
    case BOX:
      if (n == 1 || n == 2)  *x += obj.kind.box.w;
      if (n == 2 || n == 3)  *y += obj.kind.box.h;
      break;
    case CIRCLE:
      if (n == 1)  *x += obj.kind.circle.dia / 2;
      break;
    case DISK:
      if (n == 1)  *x += obj.kind.disk.dia / 2;
      break;
    case LINE:
      if (n == 1)  *x += obj.kind.line.dx;
      if (n == 1)  *y += obj.kind.line.dy;
      break;
    case OVAL:
      if (n == 1 || n == 2)  *x += obj.kind.oval.w;
      if (n == 2 || n == 3)  *y += obj.kind.oval.h;
      break;
    case RULE:
      if (n == 1 || n == 2)  *x += obj.kind.rule.w;
      if (n == 2 || n == 3)  *y += obj.kind.rule.h;
      break;
    case VECTOR:
      if (n == 1)  *x += obj.kind.vector.dx;
      if (n == 1)  *y += obj.kind.vector.dy;
      break;
  }
}




/* --------------------------------------------------------------------------
   obj_Gravitate:
     If point (x, y) is within "gravity" distance of "obj", gravitates
     (x, y) to the object and returns TRUE.  Otherwise returns FALSE.
---------------------------------------------------------------------------*/
bool
obj_Gravitate(x, y, obj)
int *x, *y;
Object obj;
{
  int n;

  if (VALID(n = obj_NearPointNo(*x, *y, obj))) {
    obj_GetCoordsOfPoint(n, x, y, obj);
    return TRUE;
  }
  return FALSE;
}




/* --------------------------------------------------------------------------
   obj_GravitateOpposite:
     If point (x, y) is within "gravity" distance of "obj", gravitates
     (x, y) to the "opposite" point of the object and returns TRUE.
     Otherwise returns FALSE.
---------------------------------------------------------------------------*/
bool
obj_GravitateOpposite(x, y, obj)
int *x, *y;
Object obj;
{
  int n;

  if (VALID(n = obj_NearPointNo(*x, *y, obj)))
    if(VALID(n = obj_OppositePointNo(n, obj.code))) {
      obj_GetCoordsOfPoint(n, x, y, obj);
      return TRUE;
    }
  return FALSE;
}




/* --------------------------------------------------------------------------
  obj_IsBounded:
    Return TRUE iff "obj" is completely within the rectangle
    (x1, y1, x2, y2).
---------------------------------------------------------------------------*/
bool
obj_IsBounded(obj, x1, y1, x2, y2)
Object obj;
int x1, y1, x2, y2;
{
  short l, r, b, t;

  SORT2(x1, x2);
  SORT2(y1, y2);
  switch (obj.code) {
    case BOX:
      l = obj.x;
      r = obj.x + obj.kind.box.w;
      b = obj.y;
      t = obj.y + obj.kind.box.h;
      break;
    case CIRCLE:
      l = obj.x - obj.kind.circle.dia / 2;
      r = obj.x + obj.kind.circle.dia / 2;
      b = obj.y - obj.kind.circle.dia / 2;
      t = obj.y + obj.kind.circle.dia / 2;
      break;
    case DISK:
      l = obj.x - obj.kind.disk.dia / 2;
      r = obj.x + obj.kind.disk.dia / 2;
      b = obj.y - obj.kind.disk.dia / 2;
      t = obj.y + obj.kind.disk.dia / 2;
      break;
    case LINE:
      l = obj.x;
      r = obj.x + obj.kind.line.dx;
      b = obj.y;
      t = obj.y + obj.kind.line.dy;
      break;
    case OVAL:
      l = obj.x;
      r = obj.x + obj.kind.oval.w;
      b = obj.y;
      t = obj.y + obj.kind.oval.h;
      break;
    case RULE:
      l = obj.x;
      r = obj.x + obj.kind.rule.w;
      b = obj.y;
      t = obj.y + obj.kind.rule.h;
      break;
    case TEXT:
      l = obj.x;
      r = obj.x + obj.kind.text.w;
      b = obj.y;
      t = obj.y + obj.kind.text.h;
      break;
    case VECTOR:
      l = obj.x;
      r = obj.x + obj.kind.vector.dx;
      b = obj.y;
      t = obj.y + obj.kind.vector.dy;
      break;
  }
  SORT2(l, r);
  SORT2(t, b);
  return (l >= x1 && r <= x2 && t >= y1 && b <= y2);
}




/* --------------------------------------------------------------------------
   obj_IsCurrent:
     Returns TRUE iff an object is in the current set.
---------------------------------------------------------------------------*/

bool
obj_IsCurrent(obj)
Object obj;
{
  return (obj.status == CURRENT);
}




/* --------------------------------------------------------------------------
   obj_IsDeleted:
     Returns TRUE iff an object has been deleted.
---------------------------------------------------------------------------*/

bool
obj_IsDeleted(obj)
Object obj;
{
  return (obj.status == DELETED);
}




/* --------------------------------------------------------------------------
   obj_IsNearPt:
     Returns TRUE iff point (x1, y1) is "near" (point (x2, y2).  (the points
     are within "gravity" distance along both axes from each other)
---------------------------------------------------------------------------*/

bool
obj_IsNearPt(x1, y1, x2, y2)
int x1, y1, x2, y2;
{
  return (MAX(ABS(x2-x1), ABS(y2-y1)) <= ABS(gravity));
}




/* --------------------------------------------------------------------------
  obj_Move:
    Move an object distance "dx" in the x-direction and distance "dy" 
    in the y-direction.
---------------------------------------------------------------------------*/
void
obj_Move(obj, dx, dy)
Object *obj;
int dx, dy;
{
    obj->x += (short)dx;
    obj->y += (short)dy;
}




/* --------------------------------------------------------------------------
  obj_MovePoint:
    Modify a geometric object.
    Point (x1, y1) remains fixed, and point (x2, y2) is point on the object's
    new perimeter.
---------------------------------------------------------------------------*/
void
obj_MovePoint(x1, y1, x2, y2, obj)
int x1, y1, x2, y2;
Object *obj;
{
  int thick = obj_ThicknessOf(*obj);

  if (obj->code == VECTOR)
    if ((x1 != obj->x || y1 != obj->y)) {
      SWAP(x1, x2);
      SWAP(y1, y2);
    }
  obj_CreateGeom(obj->code, x1, y1, x2, y2, obj);
  if (VALID(thick)) 
    obj_SetThickness(thick, obj);
}




/* --------------------------------------------------------------------------
   obj_NearPointNo:
     Returns the point number (if any) which is within "gravity" distance
     from point (x, y).
---------------------------------------------------------------------------*/
int
obj_NearPointNo(x, y, obj)
int x, y;
Object obj;
{
  int a = obj.x;
  int b = obj.y;
  int dx, dy, w, h, r1, r2;

  if (obj_IsDeleted(obj)) return NONE;

  switch (obj.code) {
    case BOX:
      w = obj.kind.box.w;
      h = obj.kind.box.h;
      break;
    case CIRCLE:
      r1 = obj.kind.circle.dia / 2;
      r2 = ROUND(LENGTH(x-a, y-b));
      break;
    case DISK:
      r1 = obj.kind.disk.dia / 2;
      r2 = ROUND(LENGTH(x-a, y-b));
      break;
    case LINE:
      dx = obj.kind.line.dx;
      dy = obj.kind.line.dy;
      break;
    case OVAL:
      w = obj.kind.oval.w;
      h = obj.kind.oval.h;
      break;
    case RULE:
      w = obj.kind.rule.w;
      h = obj.kind.rule.h;
      break;
    case VECTOR:
      dx = obj.kind.vector.dx;
      dy = obj.kind.vector.dy;
      break;
  }

  switch (obj.code) {
    case BOX:
    case OVAL:
    case RULE:
      if      (obj_IsNearPt(x, y, a,   b  )) return 0;
      else if (obj_IsNearPt(x, y, a+w, b  )) return 1;
      else if (obj_IsNearPt(x, y, a+w, b+h)) return 2;
      else if (obj_IsNearPt(x, y, a,   b+h)) return 3;
      else return NONE;
    case CIRCLE:
    case DISK:
      if (obj_IsNearPt( x, y,  a, b)) return 0;
      if (obj_IsNearPt(r1, b, r2, b)) return 1;
      else return NONE;
    case LINE:
    case VECTOR:
      if      (obj_IsNearPt(x, y, a,    b   )) return 0;
      else if (obj_IsNearPt(x, y, a+dx, b+dy)) return 1;
      else return NONE;
    default:
      return NONE;
  }
}




/* --------------------------------------------------------------------------
   obj_OppositePointNo:
     Returns the point number (if any) which is "opposite" point "n".
---------------------------------------------------------------------------*/
int
obj_OppositePointNo(n, shape)
int n, shape;
{
  switch (shape) {
    case BOX:
    case OVAL:
    case RULE:
      return ((n + 2) % 4);
    case CIRCLE:
    case DISK:
      return (n = 1 ? 0 : NONE);
    case LINE:
    case VECTOR:
      return ((n + 1) % 2);
    default:
      return NONE;
  }
}



/* --------------------------------------------------------------------------
   obj_SetCurrent:
     Add an object to the current set.
---------------------------------------------------------------------------*/

void
obj_SetCurrent(obj)
Object *obj;
{
  obj->status = CURRENT;
}




/* --------------------------------------------------------------------------
   obj_SetThickness:
---------------------------------------------------------------------------*/
void
obj_SetThickness(thick, obj)
int thick;
Object *obj;
{
  switch (obj->code) {
    case BOX:    obj->kind.box.thick    = thick;  break;
    case CIRCLE: obj->kind.circle.thick = thick;  break;
    case LINE:   obj->kind.line.thick   = thick;  break;
    case OVAL:   obj->kind.oval.thick   = thick;  break;
    case VECTOR: obj->kind.vector.thick = thick;  break;
  }
}




/* --------------------------------------------------------------------------
   obj_ShapeOf:
     Returns the "shape" of an object.
---------------------------------------------------------------------------*/
int
obj_ShapeOf(obj)
Object obj;
{
  return obj.code;
}




/* --------------------------------------------------------------------------
   obj_ThicknessOf:
---------------------------------------------------------------------------*/
int
obj_ThicknessOf(obj)
Object obj;
{
  switch (obj.code) {
    case BOX:    return obj.kind.box.thick;
    case CIRCLE: return obj.kind.circle.thick;
    case LINE:   return obj.kind.line.thick;
    case OVAL:   return obj.kind.oval.thick;
    case VECTOR: return obj.kind.vector.thick;
    default:     return NONE;
  }
}




/* --------------------------------------------------------------------------
   obj_ToggleGravity:
---------------------------------------------------------------------------*/
void
obj_ToggleGravity()
{
  if (gravity) {
    gravity = 0;
    wind_UnselectMenu(GRAVITY);
  }
  else {
    gravity = 3;
    wind_SelectMenu(GRAVITY);
  }
}


/* --------------------------------------------------------------------------
   obj_ToggleGrid:
---------------------------------------------------------------------------*/
void
obj_ToggleGrid()
{
  if (grid) {
    grid = 0;
    wind_UnselectMenu(GRID);
    XSetWindowBackgroundPixmap(display, picWind, None);
    XSetWindowBackground(display, picWind, back_pixel);
  }
  else {
    grid = 1;
    wind_SelectMenu(GRID);
    XSetWindowBackgroundPixmap(display, picWind, gridTile);
  }
  XClearWindow(display, picWind);
  objs_DrawAll();
}

void obj_SnapToGrid()
{
  if (snapToGrid) {
    wind_UnselectMenu(SNAPGRID);
    snapToGrid = 0;
  }
  else {
    wind_SelectMenu(SNAPGRID);
    snapToGrid = 1;
  }
}

void obj_ResizeGrid()
{
  int num = objs_GetGridSize();

  if (num > 0) {
    gridSize = num;
    obj_SetGridTile();
  }
}

void obj_SetGridTile()
{
  int x, y, yOffset;
  XPoint dots[5];
  int size;

  if (gridTile != None) XFreePixmap(display, gridTile);
  gridTile = XCreatePixmap(display, picWind,
		16*gridSize, 16*gridSize, DefaultDepth(display, screen));

  if (gridSize >= 12) {
    dots[1].x = gridSize / 3;
    dots[1].y = 0;
    dots[2].x = gridSize - gridSize / 3 * 2;
    dots[2].y = 0;
    dots[3].x = gridSize / 3 - gridSize;
    dots[3].y = gridSize / 3;
    dots[4].x = 0;
    dots[4].y = gridSize - gridSize / 3 * 2;
    size = 5;
  } else if (gridSize >= 6) {
    dots[1].x = gridSize / 2;
    dots[1].y = 0;
    dots[2].x = - gridSize / 2;
    dots[2].y = gridSize - gridSize / 2;
    size = 3;
  } else {
    size = 1;
  }

  yOffset = (picHeight + gridSize - 1) % gridSize;

  for (x = -1; x < 16; x++) {
    for (y = -1; y < 16; y++) {
      dots[0].x = x * gridSize;
      dots[0].y = y * gridSize + yOffset;
      XDrawPoints(display, gridTile, gridGC, dots, size, CoordModePrevious);
    }
  }

  if (grid) {
    XSetWindowBackgroundPixmap(display, picWind, gridTile);
    XClearWindow(display, picWind);
    objs_DrawAll();
  }
}

/* --------------------------------------------------------------------------
   obj_ToggleThickness:
---------------------------------------------------------------------------*/
void
obj_ToggleThickness()
{
  if (thickness != 0) {
    thickness = 0;
    wind_UnselectMenu(THICK);
  }
  else {
    thickness = 2;
    wind_SelectMenu(THICK);
  }
}


/* --------------------------------------------------------------------------
   obj_UnsetCurrent:
     Remove an object from the current set.
---------------------------------------------------------------------------*/

void
obj_UnsetCurrent(obj)
Object *obj;
{
  obj->status = STATIC;
}


char *filterSpecialChars (text)
  char *text;
{
  static char newtext[STRBUFSIZE];
  char *tptr, *nptr;

  tptr = text;
  nptr = newtext;

  while (*tptr) {
    switch (*tptr) {
    case '$':
    case '&':
    case '%':
    case '#':
    case '_':
    case '{':
    case '}':
      *nptr++ = '\\';
    }
    *nptr++ = *tptr++;
  }
  *nptr = '\0';
  return(newtext);
}
      
/* --------------------------------------------------------------------------
   obj_WriteToFile:
     Output LATEX commands for an object to a file.
---------------------------------------------------------------------------*/

void
obj_WriteToFile(file, obj)
FILE *file;
Object obj;
{
  double x, y, w, h, len, xofs, yofs;
  char *filterSpecialChars();
  int dia;

  switch (obj.code) {
    case BOX:
      x = scale * obj.x;
      y = scale * obj.y;
      w = scale * obj.kind.box.w;
      h = scale * obj.kind.box.h;
      fprintf(file, "\\put");
      fprintf(file, "(%6.1f,%6.1f)", x, y);
      fprintf(file, "{\\framebox");
      fprintf(file, "(%6.1f,%6.1f)", w, h);
      fprintf(file, "{}}\n");
      break;
    case CIRCLE:
      x = scale * obj.x;
      y = scale * obj.y;
      dia = ROUND(scale * obj.kind.circle.dia);
      fprintf(file, "\\put");
      fprintf(file, "(%6.1f,%6.1f)", x, y);
      fprintf(file, "{\\circle");
      fprintf(file, "{%2d}", dia);
      fprintf(file, "}\n");
      break;
    case DISK:
      x = scale * obj.x;
      y = scale * obj.y;
      dia = ROUND(scale * obj.kind.disk.dia);
      fprintf(file, "\\put");
      fprintf(file, "(%6.1f,%6.1f)", x, y);
      fprintf(file, "{\\circle*");
      fprintf(file, "{%2d}", dia);
      fprintf(file, "}\n");
      break;
    case LINE:
      x = scale * obj.x;
      y = scale * obj.y;
      len = scale * obj.kind.line.len;
      fprintf(file, "\\put");
      fprintf(file, "(%6.1f,%6.1f)", x, y);
      fprintf(file, "{\\line");
      fprintf(file, "(%2d,%2d)", obj.kind.line.tx, obj.kind.line.ty);
      fprintf(file, "{%6.1f}", len);
      fprintf(file, "}\n");
      break;
    case OVAL:
      x = scale * obj.x;
      y = scale * obj.y;
      w = scale * obj.kind.oval.w;
      h = scale * obj.kind.oval.h;
      x += w / 2;
      y += h / 2;
      fprintf(file, "\\put");
      fprintf(file, "(%6.1f,%6.1f)", x, y);
      fprintf(file, "{\\oval");
      fprintf(file, "(%6.1f,%6.1f)", w, h);
      fprintf(file, "}\n");
      break;
    case RULE:
      x = scale * obj.x;
      y = scale * obj.y;
      w = scale * obj.kind.rule.w;
      h = scale * obj.kind.rule.h;
      fprintf(file, "\\put");
      fprintf(file, "(%6.1f,%6.1f)", x, y);
      fprintf(file, "{\\rule");
      fprintf(file, "{%6.1f\\unitlength}{%6.1f\\unitlength}", w, h);
      fprintf(file, "}\n");
      break;
    case TEXT:
      wind_GetTexOffsets(obj.kind.text.fontNum, &xofs, &yofs);
      x = xofs + scale * obj.x;
      y = yofs + scale * obj.y;
      fprintf(file, "\\put");
      fprintf(file, "(%6.1f,%6.1f)", x, y);
      fprintf(file, "{%s}\n", filterSpecialChars(obj.kind.text.txtptr));
      break;
    case VECTOR:
      x = scale * obj.x;
      y = scale * obj.y;
      len = scale * obj.kind.vector.len;
      fprintf(file, "\\put");
      fprintf(file, "(%6.1f,%6.1f)", x, y);
      fprintf(file, "{\\vector");
      fprintf(file, "(%2d,%2d)", obj.kind.vector.tx, obj.kind.vector.ty);
      fprintf(file, "{%6.1f}", len);
      fprintf(file, "}\n");
      break;
  }
}



