/*----------------------------------------------------------------------------
**   Copyright (C) 1987 Thomas Taylor
**
** This file is part of TexDraw
**
** TexDraw is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY.  No author or distributor
** accepts responsibility to anyone for the consequences of using it
** or for whether it serves any particular purpose or works at all,
** unless he says so in writing. 
**
** This software is free of charge.  It may not be re-sold.
** Everyone is granted permission to copy, modify and redistribute
** TexDraw, as long as the original copyright and this notice are
** retained.
**
**
**         Module:  texobjs.c
**
**   Program Name:  texdraw
**    Program Use:  Graphics Program for LaTeX Text Formatter
**
**         Author:  Tom Taylor
**           Date:  August 21, 1987
**
**          Class:  C Sci 498, Senior Project
**        Sponsor:  Hank Levy
**
**
**    This module contains object set functions for the texdraw program.
**
----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------
**   Copyright (C) 1988 Hiroto Kagotani
----------------------------------------------------------------------------*/


#include "texdraw.h"


#define NAMELEN    120


static int moveObj = NONE;
static int nGrav = NONE;

static int numDels = 0;
static int numCurr = 0;


static bool alignH = FALSE;
static bool alignV = FALSE;

extern int grid;

extern Vertex *poly;
extern int vertices;

extern char lineBuffer[];
extern int currentPos;

/* --------------------------------------------------------------------------
   Declare the 'objects' array.
---------------------------------------------------------------------------*/

#define MAXOBJS 1000

Object objects[MAXOBJS];
int numObjs = 0;




/* --------------------------------------------------------------------------
   objs_AddGeom:
     Add a geometric object to the set of objects.
---------------------------------------------------------------------------*/

void
objs_AddGeom(cmd, x1, y1, x2, y2)
int cmd, x1, y1, x2, y2;
{
  if (numObjs >= MAXOBJS) {
    wind_ErrorMsg1("too many objects");
    return;
  }
  obj_CreateGeom(cmd, x1, y1, x2, y2, &objects[numObjs++]);
  numCurr++;
}




/* --------------------------------------------------------------------------
   objs_AddText:
     Add a text string to the set of objects.
---------------------------------------------------------------------------*/

void
objs_AddText(cmd, x, y, str, len, font)
int cmd, x, y;
char *str;
int len;
XFontStruct *font;
{
  /* do not save an empty string */
  if (!len) return;

  /* Do not purge or unset current if cmd == READ */
  if (cmd == TEXT) {
    objs_PurgeDeletions();
    objs_UnsetCurrent();
  }

  /* see if there is room in the set of objects for the new string */
  if (numObjs >= MAXOBJS) {
    wind_ErrorMsg1("too many objects");
    return;
  }

  /* add the string to the set of objects */
  obj_CreateText(x, y, str, len, font, &objects[numObjs++]);
  numCurr++;

  /* Check if the string is inside the picture. */
  if (!obj_IsBounded(objects[numObjs-1], 0, 0, picWidth, picHeight)) {
    wind_ErrorMsg1("error - attempt to position text outside of picture");
    objs_Undo();
    return;
  }

  /* Update the display. */
  obj_DrawObj(objects[numObjs-1]);
}




/* --------------------------------------------------------------------------
   objs_Align:
     Moves a point the shortest distance possible in the x- and/or 
     y-direction to align it with some other object.
---------------------------------------------------------------------------*/
void
objs_Align(x, y)
int *x, *y;
{
  int i, a, b;
  bool aligned = FALSE;
  Object obj;

  if (!alignH && !alignV) return;

  for (i=0; i<numObjs; i++) {
    obj = objects[i];
    if (!obj_IsDeleted(obj)) {
      if (!aligned) {
        aligned = TRUE;
        a = obj_ClosestX(*x, obj);
        b = obj_ClosestY(*y, obj);
      }
      else {
        a = CLOSER(a, obj_ClosestX(*x, obj), *x);
        b = CLOSER(b, obj_ClosestY(*y, obj), *y);
      }
    }
  }
  if (aligned) {
    if (alignV) *x = a;
    if (alignH) *y = b;
  }
}




/* --------------------------------------------------------------------------
   objs_AreBoundedBy:
---------------------------------------------------------------------------*/
bool
objs_AreBoundedBy(x1, y1, x2, y2)
int x1, y1, x2, y2;
{
  int i;

  for (i=0; i<numObjs; i++) 
    if (!obj_IsDeleted(objects[i]))
      if (!obj_IsBounded(objects[i], x1, y1, x2, y2))
        return FALSE;
  return TRUE;
}




/* --------------------------------------------------------------------------
   objs_ConditionalRedraw:
     Erase all of the "deleted" objects.
---------------------------------------------------------------------------*/

void
objs_ConditionalRedraw()
{
  if (numDels <= 1) {
    objs_EraseDeletions();
    objs_DrawCurrent();
  }
  else {
    XClearWindow(display, picWind);
    objs_DrawAll();
  }
}




/* --------------------------------------------------------------------------
   Delete all objects from the set.
---------------------------------------------------------------------------*/

void
objs_DeleteAll()
{
  int i;

  wind_SelectMenu(CLEAR);
  if (numObjs == numDels) {
    wind_ErrorMsg1("nothing to delete");
  }
  else {
    XClearWindow(display, picWind);
    objs_PurgeDeletions();
    for (i=0; i<numObjs; i++) 
      obj_Delete(&objects[i]);
  }
  numDels = numObjs;
  numCurr = 0;
  wind_UnselectMenu(CLEAR);
}




/* --------------------------------------------------------------------------
   Delete objects which are in the 'current set'.
---------------------------------------------------------------------------*/

void
objs_DeleteCurrent()
{
  int i;

  wind_SelectMenu(DELETECURRENT);
  if (numCurr == 0)
    wind_ErrorMsg1("current set is empty");
  else {
    objs_PurgeDeletions();
    for (i=0; i<numObjs; i++) {
      if (numCurr == 0) break;
      if (obj_IsCurrent(objects[i])) {
        obj_Delete(&objects[i]);
        numDels++;
        numCurr--;
      }
    }
    XClearWindow(display, picWind);
    objs_DrawAll();
  }
  wind_UnselectMenu(DELETECURRENT);
}




/* --------------------------------------------------------------------------
   Draw all objects in the set.
---------------------------------------------------------------------------*/

void
objs_DrawAll()
{
  int i;

  for (i=0; i<numObjs; i++) {
    if (!obj_IsDeleted(objects[i]))
      obj_DrawObj(objects[i]);
  }
}

objs_GetGridSize()
{
  static char prompt[] = "grid size (pixels): ";
  int x = 10 + XTextWidth(textFont, prompt, str_Length(prompt));
  int y = 15;
  int num = 0;

  wind_SelectMenu(GRIDSIZE);
  wind_Message1(prompt);

  *lineBuffer = '\0';
  currentPos = 0;
  if (!driv_EnterCommand(x, y, lineBuffer, LINEBUFSIZE))
    wind_ErrorMsg1("aborted");
  else if (str_Length(lineBuffer) == 0)
    wind_ErrorMsg1("invalid number");
  else {
    read_GetNum(&num);
    if (num <= 1) {
      wind_ErrorMsg1("Cannot set grid size to <= 1");
      num = 0;
    }
    else {
      wind_Message1("");
    }
  }
  wind_UnselectMenu(GRIDSIZE);
  return(num);
}

/* --------------------------------------------------------------------------
   Draw objects in the "current set".
---------------------------------------------------------------------------*/

void
objs_DrawCurrent()
{
  int i;

  for (i=0; i<numObjs; i++) {
    if (obj_IsCurrent(objects[i]))
      obj_DrawObj(objects[i]);
  }
}

#ifdef DUMPDATA
/* --------------------------------------------------------------------------
   Dump the contents of the entire set of objects (for debug purposes).
---------------------------------------------------------------------------*/

void
objs_DumpSet()
{
  int i;

  wind_SelectMenu(DUMPDATA);

  /* Wait for the menu to be selected before starting the command. */
  /* This makes a difference with "DUMPDATA" where printf responds */
  /* much sooner than PaintMenu.                                   */
    while (XPending(display));

  printf("OBJECTS ARRAY: \n");
  for (i=0; i<numObjs; i++) {
    printf("\n");
    printf("%3d.  ", i);
    obj_DumpObj(objects[i]);
  }
  printf("numObjs=%d  numCurr=%d  numDels=%d \n", numObjs, numCurr, numDels);
  wind_UnselectMenu(DUMPDATA);
}
#endif




/* --------------------------------------------------------------------------
   Add a duplicate copy of all objects in the current set.
---------------------------------------------------------------------------*/

void
objs_DuplicateCurrent()
{
  int i = 0;
  int n;

  objs_PurgeDeletions();
  n = numObjs;
  if (numObjs + numCurr > MAXOBJS)
    wind_ErrorMsg1("too many objects");
  else
    while (i < n) {
      if (obj_IsCurrent(objects[i])) {
        obj_Copy(objects[i], &objects[numObjs]);
        obj_UnsetCurrent(&objects[i]);
        numObjs++;
      }
      i++;
    }
}




/* --------------------------------------------------------------------------
   obj_EraseDeletions:
     Erase all of the "deleted" objects.
---------------------------------------------------------------------------*/

void
objs_EraseDeletions()
{
  int i;

  for (i=0; i<numObjs; i++)
    if (obj_IsDeleted(objects[i]))
      obj_EraseObj(objects[i]);
}




/* --------------------------------------------------------------------------
   objs_Gravitate:
     "Gravitates" point (x, y) to the first object found (if any) which
     is within distance "gravity".
---------------------------------------------------------------------------*/
void
objs_Gravitate(x, y)
int *x, *y;
{
  int i;

  nGrav = NONE;
  for (i=0; i<numObjs; i++)
    if (obj_Gravitate(x, y, objects[i])) {
      nGrav = i;
      break;
    }
}


objs_SnapGrid(x, y)
  int *x, *y;
{
  int snap_x = *x % gridSize;
  int snap_y = *y % gridSize;

  *x = (snap_x < (gridSize / 2)) ? (*x - snap_x) : (*x + gridSize - snap_x);
  *y = (snap_y < (gridSize / 2)) ? (*y - snap_y) : (*y + gridSize - snap_y);
}

/* --------------------------------------------------------------------------
   objs_GravitateOpposite:
     "Gravitates" point (x, y) to the "opposite" point of the first object
     found (if any) which is within distance "gravity".
---------------------------------------------------------------------------*/
void
objs_GravitateOpposite(x, y)
int *x, *y;
{
  int i;

  nGrav = NONE;
  for (i=0; i<numObjs; i++)
    if (obj_GravitateOpposite(x, y, objects[i])) {
      nGrav = i;
      break;
    }
}




/* --------------------------------------------------------------------------
   objs_InitShape:
     Initialize shape, x1, and y1 for a command.
---------------------------------------------------------------------------*/

void
objs_InitShape(mode, shape, x1, y1, x2, y2)
int mode, *shape, *x1, *y1, x2, y2;
{
  *x1 = x2;
  *y1 = y2;
  *shape = mode;

  if(mode == RESIZE) {
    *x1 = 0;
    *y1 = 0;
  }
  else if (mode == MOVEPOINT) {
    objs_GravitateOpposite(x1, y1);
    moveObj = nGrav;
    *shape = VALID(nGrav) ? obj_ShapeOf(objects[nGrav]) : NONE;
  }
}




/* --------------------------------------------------------------------------
  objs_MoveCurrent:
    For each object in the current set, move the object distance "dx" in 
  the x-direction and distance "dy" in the y-direction.
---------------------------------------------------------------------------*/

void
objs_MoveCurrent(x1, y1, x2, y2)
int x1, y1, x2, y2;
{
  int dx = x2 - x1;
  int dy = y2 - y1;
  int i, n;

  objs_PurgeDeletions();
  n = numObjs;
  if (numObjs + numCurr > MAXOBJS)
    wind_ErrorMsg1("too many objects");
  else
    for (i=0; i<n; i++)
      if (obj_IsCurrent(objects[i])) {
        obj_Copy(objects[i], &objects[numObjs]);
        obj_Move(&objects[numObjs], dx, dy);
        obj_Delete(&objects[i]);
        numDels++;
        numObjs++;
      }
}




/* --------------------------------------------------------------------------
   objs_MovePoint:
---------------------------------------------------------------------------*/
void
objs_MovePoint(x1, y1, x2, y2)
int x1, y1, x2, y2;
{
  int i = moveObj;;

  objs_PurgeDeletions();
  if (numObjs >= MAXOBJS) {
    wind_ErrorMsg1("too many objects");
    return;
  }
  objs_UnsetCurrent();
  obj_Copy(objects[i], &objects[numObjs]);
  obj_Delete(&objects[numObjs++]);
  obj_MovePoint(x1, y1, x2, y2, &objects[i]);
  obj_SetCurrent(&objects[i]);
  numCurr++;
  numDels++;
}




/* --------------------------------------------------------------------------
   objs_PurgeDeletions:
     Destroy and remove all objects which are 'deleted'.
---------------------------------------------------------------------------*/

void
objs_PurgeDeletions()
{
  int i = 0;

  while (i < numObjs) {
    if (numDels == 0) break;
    if (obj_IsDeleted(objects[i])) {
      obj_Destroy(&objects[i]);
      objects[i] = objects[--numObjs];
      numDels--;
    }
    else
      i++;
  }
}




/* --------------------------------------------------------------------------
   objs_ReadFromFile:
     Input LATEX commands for a set of objects from a file.
---------------------------------------------------------------------------*/
void
objs_ReadFromFile()
{
  FILE *fopen(), *infile;
  static char prompt[] = "read from file: ";
  int x = 10 + XTextWidth(textFont, prompt, str_Length(prompt));
  int y = 15;
  char fileName[NAMELEN];
  int tmpThick = thickness;
  int tmpFontNum = currFontNum;

  wind_SelectMenu(READ);
  wind_Message1(prompt);

  /* Enter the name of the input file. */
  if(!driv_EnterCommand(x, y, fileName, NAMELEN))
    wind_ErrorMsg2(prompt, "aborted");

  else if (str_Length(fileName) == 0)
    wind_ErrorMsg1("invalid file name");

  else if ((infile = fopen(fileName, "r")) == NULL)
    wind_ErrorMsg2("can't open file: ", fileName);

  else {
    objs_PurgeDeletions();
    objs_UnsetCurrent();
    thickness = 0;
    while (read_ParseLine(infile));
    thickness = tmpThick;
    wind_SetPicFont(tmpFontNum);
    fclose(infile);
    wind_Message2("done reading file:  ", fileName);
  }

  wind_UnselectMenu(READ);
}




/* --------------------------------------------------------------------------
   objs_Redraw:
---------------------------------------------------------------------------*/
void
objs_Redraw()
{
  wind_SelectMenu(REDRAW);
  XClearWindow(display, picWind);
  objs_DrawAll();
  wind_UnselectMenu(REDRAW);
}




/* --------------------------------------------------------------------------
  objs_SetCurrentIfBounded:
    For each object in the set, if the object is completely within the 
    rectangle (x1, y1, x2, y2) add it to the current set, otherwise remove
    it from the current set.
---------------------------------------------------------------------------*/

void
objs_SetCurrentIfBounded(x1, y1, x2, y2)
int x1, y1, x2, y2;
{
  int i;

  numCurr = 0;
  for (i= 0; i<numObjs; i++) 
    if (!obj_IsDeleted(objects[i]))
      if (obj_IsBounded(objects[i], x1, y1, x2, y2)) {
        obj_SetCurrent(&objects[i]);
        numCurr++;
      }
      else {
        obj_UnsetCurrent(&objects[i]);
      }
}




/* --------------------------------------------------------------------------
    objs_ShowCurrent:
    1.  Clear the picture window.
    2.  Draw objects in the "current set".
---------------------------------------------------------------------------*/

void
objs_ShowCurrent()
{
  wind_SelectMenu(SHOWCURRENT);
  if (numCurr == 0)
    wind_ErrorMsg1("the current set is empty");
  else {
    XClearWindow(display, picWind);
    objs_DrawCurrent();
  }
  wind_UnselectMenu(SHOWCURRENT);
}




/* --------------------------------------------------------------------------
   objs_SwapCurrentWithDeletions:
     Swap objects in the 'current set' and the 'deleted set'.
---------------------------------------------------------------------------*/

void
objs_SwapCurrentWithDeletions()
{
  int i;

  for (i=0; i<numObjs; i++)
    if (obj_IsDeleted(objects[i]))  obj_SetCurrent(&objects[i]);
    else if (obj_IsCurrent(objects[i]))  obj_Delete(&objects[i]);

  SWAP(numDels, numCurr);
}




/* --------------------------------------------------------------------------
   objs_ToggleHoriz:
---------------------------------------------------------------------------*/
void
objs_ToggleHoriz()
{
  if (alignH) {
    alignH = FALSE;
    wind_UnselectMenu(ALIGNH);
  }
  else {
    alignH = TRUE;
    wind_SelectMenu(ALIGNH);
  }
}




/* --------------------------------------------------------------------------
   objs_ToggleVert:
---------------------------------------------------------------------------*/
void
objs_ToggleVert()
{
  if (alignV) {
    alignV = FALSE;
    wind_UnselectMenu(ALIGNV);
  }
  else {
    alignV = TRUE;
    wind_SelectMenu(ALIGNV);
  }
}




/* --------------------------------------------------------------------------
   objs_Undo:
     1.  Swap objects in the 'current set' and the 'deleted set'.
     2.  Clear the picture window.
     3.  Redraw the set of objects.
---------------------------------------------------------------------------*/

void
objs_Undo()
{
  int i;

  wind_SelectMenu(UNDO);
  for (i=0; i<numObjs; i++)
    if (obj_IsDeleted(objects[i]))  obj_SetCurrent(&objects[i]);
    else if (obj_IsCurrent(objects[i]))  obj_Delete(&objects[i]);
  XClearWindow(display, picWind);
  objs_DrawAll();
  SWAP(numDels, numCurr);
  wind_UnselectMenu(UNDO);
}




/* --------------------------------------------------------------------------
   objs_UnsetCurrent:
     Make the 'current set' be empty.
---------------------------------------------------------------------------*/

void
objs_UnsetCurrent()
{
  int i;

  for (i=0; i<numObjs; i++) {
    if (numCurr == 0) break;
    if (obj_IsCurrent(objects[i])) {
      obj_UnsetCurrent(&objects[i]);
      numCurr--;
    }
  }
}




/* --------------------------------------------------------------------------
   objs_Update:
     Update the set of objects and the display.
---------------------------------------------------------------------------*/
void
objs_Update(mode, shape, x1, y1, x2, y2)
int mode, shape, x1, y1, x2, y2;
{
  int w, h;

#if FALSE
  printf("   Update: mode=%3d  shape=%3d \n", mode, shape);
  printf("             x1=%3d     x2=%3d \n", x1, x2);
  printf("             y1=%3d     y2=%3d \n\n", y1, y2);
#endif

  /* Do not save zero-dimension objects, make zero-distance moves, */
  /* or create an identical copy of the current set.               */
  switch (shape) {
    case CIRCLE:
    case DISK:
    case LINE:
    case VECTOR:
    case COPYCURRENT:
    case MOVECURRENT:
      if (x1 == x2 && y1 == y2) return;
      break;
    case BOX:
    case OVAL:
    case RULE:
      if (MIN(ABS(x2-x1), ABS(y2-y1)) < 2) return;
      break;
  }

  /* Update the set of objects. */
  switch (mode) {
    case BOX:
    case CIRCLE:
    case DISK:
    case LINE:
    case OVAL:
    case RULE:
    case VECTOR:
      objs_PurgeDeletions();
      objs_UnsetCurrent();
    case READ:
      objs_AddGeom(shape, x1, y1, x2, y2);
      break;

    case COPYCURRENT:
      objs_DuplicateCurrent();
    case MOVECURRENT:
      objs_MoveCurrent(x1, y1, x2, y2);
      break;

    case MOVEPOINT:
      objs_MovePoint(x1, y1, x2, y2);
      break;

    case SELECTCURRENT:
      objs_SetCurrentIfBounded(x1, y1, x2, y2);
      break;

    case RESIZE:
      objs_PurgeDeletions();
      objs_UnsetCurrent();
      break;
  }

  /* Check if all objects are inside the picture. */
  w = (mode == RESIZE) ? x2 : picWidth;
  h = (mode == RESIZE) ? y2 : picHeight;
  if (!objs_AreBoundedBy(0, 0, w, h)) {
    wind_ErrorMsg1("error - attempt to position object(s) outside of picture");
    objs_Undo();
    return;
  }

  /* Update the display. */
  switch (mode) {
    Object obj;
    case BOX:
    case CIRCLE:
    case DISK:
    case LINE:
    case OVAL:
    case RULE:
    case VECTOR:
    case READ:
      obj_CreateGeom(shape, x1, y1, x2, y2, &obj);
      switch (shape) {
	case BOX:
	  poly_Box(obj, thickness ? thickGC : thinGC);
	  break;
	case CIRCLE:
	  poly_Circle(obj, thickness ? thickGC : thinGC);
	  break;
	case DISK:
	  poly_Disk(obj, tiledGC);
	  break;
	case LINE:
	  poly_Line(obj, thickness ? thickGC : thinGC);
	  break;
	case OVAL:
	  poly_Oval(obj, thickness ? thickGC : thinGC);
	  break;
	case RULE:
	  poly_Rule(obj, tiledGC);
	  break;
	case VECTOR:
	  poly_Vector(obj, thickness ? thickGC : thinGC);
	  break;
      }
      break;
    case MOVECURRENT:
    case MOVEPOINT:
      XClearWindow(display, picWind);
      objs_DrawAll();
      break;
    case COPYCURRENT:
      objs_DrawCurrent();
      break;
    case RESIZE:
      wind_ResizePicture(w, h);
      objs_DrawAll();
      break;
  }
}




/* --------------------------------------------------------------------------
   objs_WriteToFile:
     Output LATEX commands for the set of objects to a file.
---------------------------------------------------------------------------*/
void
objs_WriteToFile()
{
  FILE *fopen(), *outfile;
  static char prompt[] = "write to file: ";
  int x = 10 + XTextWidth(textFont, prompt, str_Length(prompt));
  int y = 15;
  int w = ROUND(scale * picWidth);
  int h = ROUND(scale * picHeight);
  int i;
  char fileName[NAMELEN];
  int tmpThick = thickness;
  int objThick;
  int tmpFontNum = currFontNum;
  XFontStruct *objFont;

  wind_SelectMenu(WRITE);
  wind_Message1(prompt);

  /* Enter the name of the output file. */
  if(!driv_EnterCommand(x, y, fileName, NAMELEN))
    wind_ErrorMsg2(prompt, "aborted");

  else if (str_Length(fileName) == 0)
    wind_ErrorMsg1("invalid file name");

  else if ((outfile = fopen(fileName, "w")) == NULL)
    wind_ErrorMsg2("can't open file: ", fileName);

  else {
    /* Output a command to begin a picture. */
    fprintf(outfile, "\\begin{picture}");
    fprintf(outfile, "(%4d,%4d)\n", w, h);

    /* Set the default line thickness. */
    thickness = 0;

    /* Reset the picture font. */
    picFont = NULL;

    for (i=0; i<numObjs; i++)
      if (!obj_IsDeleted(objects[i])) {

        /* Output a command to change the line thickness if necessary. */
        if (VALID(objThick = obj_ThicknessOf(objects[i])))
          if (thickness != objThick) {
            thickness = objThick;
            if (thickness == 0) fprintf(outfile, "\\thinlines\n");
            else                fprintf(outfile, "\\thicklines\n");
          }

        /* Output a command to change the font if necessary. */
        if ((objFont = obj_FontOf(objects[i])) != NULL)
          if (picFont != objFont) {
            wind_SetPicFont(wind_LookupFontNum(objFont));
            fprintf(outfile, "\\%s\\tt\n", wind_LookupTexFromFont(picFont));
          }

        /* Output the command string for the next object. */
        obj_WriteToFile(outfile, objects[i]);
      }

    /* Restore the line thickness and font. */
    thickness = tmpThick;
    wind_SetPicFont(tmpFontNum);

    /* Output a command to end the picture, and close the file. */
    fprintf(outfile, "\\end{picture}\n");
    fclose(outfile);
    wind_Message2("done writing file:  ", fileName);
  }

  wind_UnselectMenu(WRITE);
}






