/*----------------------------------------------------------------------------
**   Copyright (C) 1987 Thomas Taylor
**
** This file is part of TexDraw
**
** TexDraw is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY.  No author or distributor
** accepts responsibility to anyone for the consequences of using it
** or for whether it serves any particular purpose or works at all,
** unless he says so in writing. 
**
** This software is free of charge.  It may not be re-sold.
** Everyone is granted permission to copy, modify and redistribute
** TexDraw, as long as the original copyright and this notice are
** retained.
**
**
**         Module:  texread.c
**
**   Program Name:  texdraw
**    Program Use:  Graphics Program for LaTeX Text Formatter
**
**         Author:  Tom Taylor
**           Date:  August 21, 1987
**
**          Class:  C Sci 498, Senior Project
**        Sponsor:  Hank Levy
**
**
**    This module contains LaTeX string parsing functions for the texdraw 
**    program.
**
----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------
**   Copyright (C) 1988 Hiroto Kagotani
----------------------------------------------------------------------------*/


#include "texdraw.h"

#define SYM_SLASH     1
#define SYM_COMMA     2
#define SYM_LPAREN    3
#define SYM_RPAREN    4
#define SYM_LBRACE    5
#define SYM_RBRACE    6
#define SYM_LBRACK    7
#define SYM_RBRACK    8
#define SYM_THIN      9
#define SYM_THICK    10
#define SYM_NUM      11
#define SYM_REAL     12
#define SYM_PUT      13
#define SYM_UNIT     14

#define SYM_BOX      21
#define SYM_CIRCLE   22
#define SYM_DASHBOX  24
#define SYM_DISK     25
#define SYM_LINE     26
#define SYM_OVAL     27
#define SYM_RULE     28
#define SYM_STRING   29
#define SYM_VECTOR   30

#define SYM_BEGIN    40
#define SYM_END      41
#define SYM_PICTURE  42

#define SYM_small    50
#define SYM_normal   51
#define SYM_large    52
#define SYM_Large    53


static bool lineError;

char lineBuffer[LINEBUFSIZE];  
int currentPos;                       /* index into the line buffer */
static char command[CMDBUFSIZE];

static int thisSym;
static int thisNum;
static double thisReal;
static char thisString[STRBUFSIZE];

double scale = 0.84315;          /* scales screen dimensions down to LaTeX */
                                 /* pts / pixel                            */

extern char prevch;

/* --------------------------------------------------------------------------
   read_GetCoords:
     Get a pair of parenthesized coordinates from the line buffer.
---------------------------------------------------------------------------*/
void 
read_GetCoords(a, b)
int *a, *b;
{
  read_MatchSym(SYM_LPAREN);
  read_GetNum(a);
  read_MatchSym(SYM_COMMA);
  read_GetNum(b);
  read_MatchSym(SYM_RPAREN);
}




/* --------------------------------------------------------------------------
   read_GetNum:
     1.  Get a number from the line buffer.
     2.  Store the symbol SYM_NUM in "thisSym".
     3.  Store the number's value in "thisNum.
---------------------------------------------------------------------------*/
void
read_GetNum(n)
int *n;
{
  int sign = 1;
  thisSym = NONE;
  thisNum = 0;

  /* gobble white space */
  while (read_IsSpace(lineBuffer[currentPos])) currentPos++;

  if (lineBuffer[currentPos] == '-') {
    sign = -1;
    currentPos++;
  }
  while (read_IsDigit(lineBuffer[currentPos])) {
    thisNum *= 10;
    thisNum += lineBuffer[currentPos++] - '0';
    thisSym = SYM_NUM;
  }
  thisNum *= sign;
  *n = thisNum;

  read_PrintSym();

}




/* --------------------------------------------------------------------------
   read_GetObject:
     1.  Get an object and its parameters from the line buffer.
     2.  Unscale the parameters, and create the object.
---------------------------------------------------------------------------*/
void
read_GetObject()
{
  double fx, fy, fw, fh, flen;
  int x, y, w, h, dia, rad, tx, ty, len, dx, dy, part;
  int thisObj;

  read_GetRealCoords(&fx, &fy);
  read_MatchSym(SYM_LBRACE);

  if ((prevch = lineBuffer[currentPos]) == '\\') {
    switch (lineBuffer[currentPos+1]) {
      case '$':
      case '&':
      case '%':
      case '#':
      case '_':
      case '{':
      case '}':
        thisObj = SYM_STRING;
        currentPos++;
        break;
      default:
        read_MatchSym(SYM_SLASH);
        read_GetSym();
        thisObj = thisSym;
    }
  }
  else {
    thisObj = SYM_STRING;
  }

  switch (thisObj) {
    case SYM_BOX:
      read_GetRealCoords(&fw, &fh);
      read_MatchSym(SYM_LBRACE);
      read_MatchSym(SYM_RBRACE);
      break;
    case SYM_CIRCLE:
    case SYM_DISK:
      read_MatchSym(SYM_LBRACE);
      read_GetNum(&dia);
      read_MatchSym(SYM_RBRACE);
      break;
    case SYM_LINE:
    case SYM_VECTOR:
      read_GetCoords(&tx, &ty);
      read_MatchSym(SYM_LBRACE);
      read_GetReal(&flen);
      read_MatchSym(SYM_RBRACE);
      break;
    case SYM_OVAL:
      read_GetRealCoords(&fw, &fh);
      read_GetOvalOption(&part);
      break;
    case SYM_RULE:
      read_MatchSym(SYM_LBRACE);
      read_GetReal(&fw);
      read_MatchSym(SYM_SLASH);
      read_MatchSym(SYM_UNIT);
      read_MatchSym(SYM_RBRACE);
      read_MatchSym(SYM_LBRACE);
      read_GetReal(&fh);
      read_MatchSym(SYM_SLASH);
      read_MatchSym(SYM_UNIT);
      read_MatchSym(SYM_RBRACE);
      break;
    case SYM_STRING:
      read_GetString();
      break;
  }

  read_MatchSym(SYM_RBRACE);

  if (!lineError) {
    if      (thisObj == SYM_BOX)     thisObj = BOX;
    else if (thisObj == SYM_CIRCLE)  thisObj = CIRCLE;
    else if (thisObj == SYM_DASHBOX) /***thisObj = DASHBOX***/;
    else if (thisObj == SYM_DISK)    thisObj = DISK;
    else if (thisObj == SYM_LINE)    thisObj = LINE;
    else if (thisObj == SYM_OVAL)    thisObj = OVAL;
    else if (thisObj == SYM_RULE)    thisObj = RULE;
    else if (thisObj == SYM_STRING)  thisObj = TEXT;
    else if (thisObj == SYM_VECTOR)  thisObj = VECTOR;

    switch(thisObj) {
      case BOX:
      case RULE:
        x = ROUND(fx / scale);
        y = ROUND(fy / scale);
        w = ROUND(fw / scale);
        h = ROUND(fh / scale);
        objs_Update(READ, thisObj, x, y, x+w, y+h);
        break;
      case CIRCLE:
      case DISK:
        x = ROUND(fx / scale);
        y = ROUND(fy / scale);
        rad = ROUND(dia / 2.0 / scale);
        objs_Update(READ, thisObj, x, y, x+rad, y);
        break;
      case LINE:
      case VECTOR:
        x = ROUND(fx / scale);
        y = ROUND(fy / scale);
        len = ROUND(flen / scale);
        ty = ty;
        geom_TexToGraphSlope(tx, ty, len, &dx, &dy);
        objs_Update(READ, thisObj, x, y, x+dx, y+dy);
        break;
      case OVAL:
        x = ROUND((fx - fw / 2.0) / scale);
        y = ROUND((fy - fh / 2.0) / scale);
        w = ROUND(fw / scale);
        h = ROUND(fh / scale);
        objs_Update(READ, OVAL, x, y, x+w, y+h);
        break;
      case TEXT:
        x = ROUND((fx - texXofs) / scale);
        y = ROUND((fy - texYofs) / scale);
        len = str_Length(thisString);
        objs_AddText(READ, x, y, thisString, len, picFont);
        break;
    }
  }
}




/* --------------------------------------------------------------------------
   read_GetOvalOption:
     1.  Get a oval option like "[tl]" and return OVAL_TL etc.
     2.  If no option is specified, return OVAL_ALL.
---------------------------------------------------------------------------*/
void
read_GetOvalOption(part)
int *part;
{
  *part = 0;

  if (lineBuffer[currentPos] == '[') {
    while (lineBuffer[++currentPos] != ']') {
      switch (lineBuffer[currentPos]) {
        case 't': *part |= OVAL_noT; break;
        case 'l': *part |= OVAL_noL; break;
        case 'b': *part |= OVAL_noB; break;
        case 'r': *part |= OVAL_noR; break;
        case '\0': lineError = 1; return;
      }
    }
    currentPos ++;
  }
}




/* --------------------------------------------------------------------------
   read_GetPicture:
     1.  Get a picture symbol and its size parameters from the input line.
     2.  Resize the picture window if necessary.
---------------------------------------------------------------------------*/
void
read_GetPicture()
{
  int w, h;

  read_MatchSym(SYM_LBRACE);
  read_MatchSym(SYM_PICTURE);
  read_MatchSym(SYM_RBRACE);
  read_GetCoords(&w, &h);
  w = ROUND(w / scale);
  h = ROUND(h / scale);
  if (!lineError) {
    w = MAX(w, picWidth);
    h = MAX(h, picHeight);
    if ((w > picWidth) || (h > picHeight)) {
      wind_ResizePicture(w, h);
      objs_DrawAll();
    }
  }
}




/* --------------------------------------------------------------------------
   read_GetReal:
     1.  Get a floating point value from the line buffer.
     2.  Store the symbol SYM_REAL in "thisSym".
     3.  Store the value in thisReal;
     4.  Return the floating-point value.
---------------------------------------------------------------------------*/
void
read_GetReal(n)
double *n;
{
  double sign = 1.0;
  double val = 0.0;
  double dec = 1.0;

  thisSym = NONE;
  thisReal = 0;

  /* gobble white space */
  while (read_IsSpace(lineBuffer[currentPos])) currentPos++;

  if (lineBuffer[currentPos] == '-') {
    sign = -1;
    currentPos++;
  }
  while (read_IsDigit(lineBuffer[currentPos])) {
    val *= 10.0;
    val += (double)(lineBuffer[currentPos++] - '0');
    thisSym = SYM_REAL;
  }
  if (lineBuffer[currentPos] == '.') {
    thisSym = SYM_REAL;
    currentPos++;
    while (read_IsDigit(lineBuffer[currentPos])) {
      dec /= 10.0;
      val += dec * (lineBuffer[currentPos++] - '0');
    }
  }

  *n = thisReal = val * sign;

  read_PrintSym();

}



/* --------------------------------------------------------------------------
   read_GetRealCoords:
     Get a pair of floating-point coordinates from the line buffer.
---------------------------------------------------------------------------*/
void
read_GetRealCoords(a, b)
double *a, *b;
{
  read_MatchSym(SYM_LPAREN);
  read_GetReal(a);
  read_MatchSym(SYM_COMMA);
  read_GetReal(b);
  read_MatchSym(SYM_RPAREN);
}




/* --------------------------------------------------------------------------
   read_GetString:
     1.  Get a text string from the line buffer.
     2.  Store the symbol SYM_STRING in "thisSym".
     3.  Store the string (null terminated) in "thisString".
---------------------------------------------------------------------------*/
void
read_GetString()
{
  int i = 0;
  char ch;

  thisSym = SYM_STRING;

  while (1) {
    ch = lineBuffer[currentPos];
    if (ch == '\\') {
      currentPos++;
    }
    else if (read_IsStringChar(ch) || (prevch == '\\')) {
      thisString[i++] = lineBuffer[currentPos++];
    }
    else {
      break;
    }
    prevch = ch;
  }
  thisString[i] = '\0';

  read_PrintSym();
}

/* --------------------------------------------------------------------------
   read_GetSym:
     1.  Get a symbol from the line buffer.
     2.  Store the symbol in "thisSym".
---------------------------------------------------------------------------*/
void
read_GetSym()
{
  int i = 0;
  thisSym = NONE;

#ifdef trace
  printf("    enter - read_GetSym() \n");
#endif

  /* gobble white space */
  while (read_IsSpace(lineBuffer[currentPos])) currentPos++;

  if (read_IsDelimiter(lineBuffer[currentPos]))
    switch (lineBuffer[currentPos++]) {
      case '\\':  thisSym = SYM_SLASH;    break;
      case ',':   thisSym = SYM_COMMA;    break;
      case '(':   thisSym = SYM_LPAREN;   break;
      case ')':   thisSym = SYM_RPAREN;   break;
      case '{':   thisSym = SYM_LBRACE;   break;
      case '}':   thisSym = SYM_RBRACE;   break;
      case '[':   thisSym = SYM_LBRACK;   break;
      case ']':   thisSym = SYM_RBRACK;   break;
    }

  else {
    while (read_IsCommandChar(lineBuffer[currentPos]))
      command[i++] = lineBuffer[currentPos++];
    command[i] = '\0';

#ifdef debug
    printf("command=%s \n", command);
#endif

    if      (str_IsEqual(command, "begin"))      thisSym = SYM_BEGIN;
    else if (str_IsEqual(command, "framebox"))   thisSym = SYM_BOX;
    else if (str_IsEqual(command, "circle"))     thisSym = SYM_CIRCLE;
    else if (str_IsEqual(command, "circle*"))    thisSym = SYM_DISK;
    else if (str_IsEqual(command, "dashbox"))    thisSym = SYM_DASHBOX;
    else if (str_IsEqual(command, "end"))        thisSym = SYM_END;
    else if (str_IsEqual(command, "large"))      thisSym = SYM_large;
    else if (str_IsEqual(command, "Large"))      thisSym = SYM_Large;
    else if (str_IsEqual(command, "line"))       thisSym = SYM_LINE;
    else if (str_IsEqual(command, "normalsize")) thisSym = SYM_normal;
    else if (str_IsEqual(command, "oval"))       thisSym = SYM_OVAL;
    else if (str_IsEqual(command, "picture"))    thisSym = SYM_PICTURE;
    else if (str_IsEqual(command, "put"))        thisSym = SYM_PUT;
    else if (str_IsEqual(command, "rule"))       thisSym = SYM_RULE;
    else if (str_IsEqual(command, "small"))      thisSym = SYM_small;
    else if (str_IsEqual(command, "thicklines")) thisSym = SYM_THICK;
    else if (str_IsEqual(command, "thinlines"))  thisSym = SYM_THIN;
    else if (str_IsEqual(command, "unitlength")) thisSym = SYM_UNIT;
    else if (str_IsEqual(command, "vector"))     thisSym = SYM_VECTOR;
  }

  read_PrintSym();

#ifdef trace
  printf("     exit - read_GetSym() \n");
#endif
}




/* --------------------------------------------------------------------------
   read_IsCommandChar:
     Returns TRUE iff character "c" is a valid character in a command name.
---------------------------------------------------------------------------*/
bool
read_IsCommandChar(c)
{
  return ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') || (c == '*'));
}



/* --------------------------------------------------------------------------
   read_IsDelimiter:
     Returns TRUE iff "c" is a LaTeX command delimiter.
---------------------------------------------------------------------------*/
bool
read_IsDelimiter(c)
char c;
{
  return (c == '(' || c == ')' ||
          c == '{' || c == '}' ||
          c == ',' || c == '\\');
}




/* --------------------------------------------------------------------------
   read_IsDigit:
     Returns TRUE iff character "c" is a numeric digit.
---------------------------------------------------------------------------*/
bool
read_IsDigit(c)
char c;
{
  return (c >= '0' && c <= '9');
}




/* --------------------------------------------------------------------------
   read_IsSpace:
     Returns TRUE iff character "c" is "white space".
---------------------------------------------------------------------------*/
bool
read_IsSpace(c)
char c;
{
  static int inComment;
  if (c == '%') inComment = 1;
  if (c == '\n') inComment = 0;
  return (inComment || c == ' ' || c == '\t' || c == '\n');
}





/* --------------------------------------------------------------------------
   read_IsStringChar:
     Returns TRUE iff "c" is a valid LaTeX string character.
---------------------------------------------------------------------------*/
bool
read_IsStringChar(c)
char c;
{
  return (c >= ' ' && c != '}');
}

bool
read_IsInputStringChar(c)
char c;
{
  return (c >= ' ' && c != '~' && c != '^' && c != '\\');
}

/* --------------------------------------------------------------------------
   read_MatchSym:
     1.  Get the next symbol from the line buffer.
     2.  Report an error if it was not the expected symbol.
---------------------------------------------------------------------------*/
void
read_MatchSym(sym)
int sym;
{
  read_GetSym();
  if (thisSym != sym)
    lineError = TRUE;
}

/* --------------------------------------------------------------------------
   read_ParseLine:
     1.  Parse a LATEX command in lineBuffer.
     2.  Iff it is a "put" command, fill in the object data.
     3.  Returns TRUE iff a line was read.
---------------------------------------------------------------------------*/
bool
read_ParseLine(infile)
FILE *infile;
{

  /* Read a line from "infile" into "lineBuffer".  Skip empty lines. */
  do {
    if (fgets(lineBuffer, LINEBUFSIZE, infile) == NULL)
      return FALSE;
  } while (str_Length(lineBuffer) == 0);

#ifdef debug
    printf("\n*** entering read_ParseLine *** \n");
    printf("%s",lineBuffer);
#endif

  lineError = FALSE;
  currentPos = 0;
  read_MatchSym(SYM_SLASH);
  read_GetSym();
  switch (thisSym) {
    case SYM_BEGIN:
      read_GetPicture();
      break;
    case SYM_END:
      /* do nothing */
      break;
    case SYM_small:
      picFont = wind_LookupFontFromTex("small");
      wind_SetPicFont(wind_LookupFontNum(picFont));
      break;
    case SYM_normal:
      picFont = wind_LookupFontFromTex("normalsize");
      wind_SetPicFont(wind_LookupFontNum(picFont));
      break;
    case SYM_large:
      picFont = wind_LookupFontFromTex("large");
      wind_SetPicFont(wind_LookupFontNum(picFont));
      break;
    case SYM_Large:
      picFont = wind_LookupFontFromTex("Large");
      wind_SetPicFont(wind_LookupFontNum(picFont));
      break;
    case SYM_THICK:
      thickness = 2;
      break;
    case SYM_THIN:
      thickness = 0;
      break;
    case SYM_PUT:
      read_GetObject();
      break;
  }

  if (lineError) {
    wind_ErrorMsg2("*** don't understand line ***", lineBuffer);
    printf("*** don't understand line *** \n");
    printf("s\n", lineBuffer);
  }

  return TRUE;
}




/* --------------------------------------------------------------------------
   read_PrintSym:
     Print the name of a symbol.  For test purposes only.
---------------------------------------------------------------------------*/
void
read_PrintSym()
{
#ifdef debug

  switch(thisSym) {
    case SYM_BOX:        printf("SYM_BOX        \n");             break;
    case SYM_BEGIN:      printf("SYM_BEGIN      \n");             break;
    case SYM_CIRCLE:     printf("SYM_CIRCLE     \n");             break;
    case SYM_DASHBOX:    printf("SYM_DASHBOX    \n");             break;
    case SYM_DISK:       printf("SYM_DISK       \n");             break;
    case SYM_COMMA:      printf("SYM_COMMA      \n");             break;
    case SYM_END:        printf("SYM_END        \n");             break;
    case SYM_large:      printf("SYM_large      \n");             break;
    case SYM_Large:      printf("SYM_Large      \n");             break;
    case SYM_LBRACE:     printf("SYM_LBRACE     \n");             break;
    case SYM_LBRACK:     printf("SYM_LBRACK     \n");             break;
    case SYM_LINE:       printf("SYM_LINE       \n");             break;
    case SYM_LPAREN:     printf("SYM_LPAREN     \n");             break;
    case SYM_normal:     printf("SYM_normal     \n");             break;
    case SYM_NUM:        printf("SYM_NUM: %d    \n", thisNum);    break;
    case SYM_OVAL:       printf("SYM_OVAL       \n");             break;
    case SYM_PICTURE:    printf("SYM_PICTURE    \n");             break;
    case SYM_PUT:        printf("SYM_PUT        \n");             break;
    case SYM_RBRACE:     printf("SYM_RBRACE     \n");             break;
    case SYM_RBRACK:     printf("SYM_RBRACK     \n");             break;
    case SYM_REAL:       printf("SYM_REAL: %g   \n", thisReal);   break;
    case SYM_RPAREN:     printf("SYM_RPAREN     \n");             break;
    case SYM_RULE:       printf("SYM_RULE       \n");             break;
    case SYM_SLASH:      printf("SYM_SLASH      \n");             break;
    case SYM_small:      printf("SYM_small      \n");             break;
    case SYM_STRING:     printf("SYM_STRING: %s \n", thisString); break;
    case SYM_THICK:      printf("SYM_THICK      \n");             break;
    case SYM_THIN:       printf("SYM_THIN       \n");             break;
    case SYM_VECTOR:     printf("SYM_VECTOR     \n");             break;
    default:          printf("unknown symbol    \n");             break;
  }
#endif
}

