/*----------------------------------------------------------------------------
**   Copyright (C) 1987 Thomas Taylor
**
** This file is part of TexDraw
**
** TexDraw is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY.  No author or distributor
** accepts responsibility to anyone for the consequences of using it
** or for whether it serves any particular purpose or works at all,
** unless he says so in writing. 
**
** This software is free of charge.  It may not be re-sold.
** Everyone is granted permission to copy, modify and redistribute
** TexDraw, as long as the original copyright and this notice are
** retained.
**
**
**         Module:  texwind.c
**
**   Program Name:  texdraw
**    Program Use:  Graphics Program for LaTeX Text Formatter
**
**         Author:  Tom Taylor
**           Date:  August 21, 1987
**
**          Class:  C Sci 498, Senior Project
**        Sponsor:  Hank Levy
**
**
**    This module contains window and display functions for the texdraw 
**    program.
**
----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------
**   Copyright (C) 1988 Hiroto Kagotani
----------------------------------------------------------------------------*/


#include "texdraw.h"
#define cursorFontName "cursor"

int currFontNum = 0;


/*----------------------------------------------------------------------------
  global (to all modules) variables:
----------------------------------------------------------------------------*/
Display *display;

int screen;

Window textWind;                                   /* the text window       */
Window picWind;                                    /* the picture window    */
Window graphWind;                                  /* the background window */

int picWidth;                    /* the current picture height              */
int picHeight;                   /* the current picture width               */

XFontStruct *menuFont;           /* the current font for the menu windows   */
XFontStruct *picFont;            /* the current font for the picture window */
XFontStruct *textFont;           /* the current font for the text window    */

int scrXofs;                            /* the current screen x-offset      */
int scrYofs;                            /* the current screen y-offset      */
double texXofs;                         /* the current LaTeX x-offset       */
double texYofs;                         /* the current LaTeX y-offset       */

Cursor smallCrossCursor;

unsigned long fore_pixel;
unsigned long back_pixel;
unsigned long brdr_pixel;



/*----------------------------------------------------------------------------
  local (to this module) variables and constants:
----------------------------------------------------------------------------*/
#define MIN_WIND   10            /* the smallest allowable window size      */

#define MENU_ROWS   ( 14 )
#define MENU_COLS   ( 3 )
#define NUM_MENUS   ( MENU_ROWS * MENU_COLS )
#define MENU_WIDTH  ( 50 )
#define MENU_HEIGHT ( 46 )
#define MENU_BORDER ( 1 )

#define TEXT_X      ( MENU_COLS * (MENU_WIDTH+MENU_BORDER) + MENU_BORDER )
#define TEXT_Y      ( GR_HEIGHT - TEXT_HEIGHT - TEXT_BORDER * 2 )
#define TEXT_WIDTH  ( 500 )
#define TEXT_HEIGHT ( 50 )
#define TEXT_BORDER ( 1 )

#define PIC_X       ( MENU_COLS * (MENU_WIDTH+MENU_BORDER) + MENU_BORDER )
#define PIC_BORDER  ( 2 )

#define GR_WIDTH    ( MENU_COLS * (MENU_WIDTH+MENU_BORDER)\
			+ MENU_BORDER + TEXT_WIDTH)
#define GR_HEIGHT   ( MENU_ROWS * (MENU_HEIGHT+MENU_BORDER) + MENU_BORDER )
#define GR_BORDER   ( 2 )

static bool menuSelected[NUM_MENUS];          /* which menus are shaded     */
static int volume;                            /* keyboard beep volume       */

double unitLength = 1 / 72.27;          /* inches per pixel           */

static Window menuWind[MENU_ROWS][MENU_COLS];




/*----------------------------------------------------------------------------
  Define the font entries.
----------------------------------------------------------------------------*/
typedef struct _FontEntry {
  XFontStruct *info;              /* X font info pointer  */
  GC writeGC;                     /* X GC number          */
  GC eraseGC;                     /* X GC number          */
  char *xName;                    /* name of the X font   */
  char *texName;                  /* name of the TeX font */
  int scrXofs;                    /* screen x-offset      */
  int scrYofs;                    /* screen y-offset      */
  double texXofs;                 /* LaTeX x-offset       */
  double texYofs;                 /* LaTeX y-offset       */
} FontEntry;


FontEntry fontEntries[] = {
  { NULL, 0, 0, "6x10", "small",      0,  0, 1.0, 0.5 },
  { NULL, 0, 0, "6x12", "normalsize", 1,  0, 1.0, 0.5 },
  { NULL, 0, 0, "8x13", "large",      0, -1, 1.0, 1.0 },
  { NULL, 0, 0, "9x15", "Large",      0,  0, 1.0, 1.0 }
};

#define NUM_FONTS (sizeof(fontEntries) / sizeof(FontEntry))



/*----------------------------------------------------------------------------
  Define the contents of the menu windows.
----------------------------------------------------------------------------*/
typedef struct _MenuEntry {
  int num;                       /* menu (command) number */
  int lines;                     /* number of menu lines  */
  char *line1;                   /* contents of line 1    */
  char *line2;                   /* contents of line 2    */
} MenuEntry;

static MenuEntry menuEntries[] = {
  { ALIGNH,         2, " ALIGN ", " HORIZ " },
  { ALIGNV,         2, " ALIGN ", " VERT  " },
  { BOX,            2, " DRAW  ", "  BOX  " },
  { CIRCLE,         2, " DRAW  ", "CIRCLE " },
  { CLEAR,          1, " CLEAR ", "       " },
  { COPYCURRENT,    2, " COPY  ", "CURRENT" },
  { DELETECURRENT,  2, "DELETE ", "CURRENT" },
  { DISK,           2, " DRAW  ", " DISK  " },
#ifdef DUMPDATA
  { DUMPDATA,       2, " DUMP  ", " DATA  " },
#endif
  { FONTS,          1, " FONT  ", "       " },        
  { GRAVITY,        1, "GRAVITY", "       " },
#ifdef GRID
  { GRID,           1, " GRID  ", "       " },
  { SNAPGRID,       2, "SNAP TO", " GRID  " },
  { GRIDSIZE,       2, " GRID  ", " SIZE  " },
#endif
  { LINE,           2, " DRAW  ", " LINE  " },
  { MOVECURRENT,    2, " MOVE  ", "CURRENT" },
  { MOVEPOINT,      2, " MOVE  ", " POINT " },
  { OVAL,           2, " DRAW  ", " OVAL  " },
  { RULE,           2, " DRAW  ", " RULE  " },
  { QUIT,           1, " QUIT  ", "       " },
  { READ,           1, " READ  ", "       " },
  { REDRAW,         1, "REDRAW ", "       " },
  { RESIZE,         2, "RESIZE ", "PICTURE" },
  { SELECTCURRENT,  2, "SELECT ", "CURRENT" },
  { SHOWCURRENT,    2, " SHOW  ", "CURRENT" },
  { TEXT,           2, " ENTER ", " TEXT  " },
  { THICK,          2, " THICK ", " LINES " },
  { UNDO,           1, " UNDO  ", "       " },
  { VECTOR,         2, " DRAW  ", "VECTOR " },
  { WRITE,          1, " WRITE ", "       " }
};

#define NUM_ENTRIES (sizeof(menuEntries) / sizeof(MenuEntry))




/*----------------------------------------------------------------------------
  wind_ChangeFont:
----------------------------------------------------------------------------*/
void
wind_ChangeFont()
{
  /* Shade the menu window. */
  wind_SelectMenu(FONTS);

  /* Advance "currFontNum" to the next non-NULL font. */
  do {
    currFontNum++;
    currFontNum %= NUM_FONTS;
    picFont = fontEntries[currFontNum].info;
  } while (picFont == NULL);

  scrXofs = fontEntries[currFontNum].scrXofs;
  scrYofs = fontEntries[currFontNum].scrYofs;
  texXofs = fontEntries[currFontNum].texXofs;
  texYofs = fontEntries[currFontNum].texYofs;

  /* Unshade the menu window, and display "FONT" */
  /* in the current picture font size.           */
  wind_UnselectMenu(FONTS);

  /* Display the name of the font in the text window. */
  wind_Message2(fontEntries[currFontNum].texName, " font selected");

}




/*----------------------------------------------------------------------------
  wind_DefinePixels:
----------------------------------------------------------------------------*/
void
wind_DefinePixels()
{
  fore_pixel = BlackPixel(display, screen);
  back_pixel = WhitePixel(display, screen);
  brdr_pixel = BlackPixel(display, screen);
}




/*----------------------------------------------------------------------------
  wind_ErrorMsg1:
    1.  Sound the bell.
    2.  Display a one-string error message in the text window.
----------------------------------------------------------------------------*/
void
wind_ErrorMsg1(str)
char str[];
{
  XBell(display, volume);
  wind_Message1(str);
}




/*----------------------------------------------------------------------------
  wind_ErrorMsg2:
    1.  Sound the bell.
    2.  Display a two-string error message in the text window.
----------------------------------------------------------------------------*/
void
wind_ErrorMsg2(str1, str2)
char str1[], str2[];
{
  XBell(display, volume);
  wind_Message2(str1, str2);
}




/*----------------------------------------------------------------------------
  wind_Finalize:
    Close the X fonts.
----------------------------------------------------------------------------*/
void
wind_Finalize()
{
  int i;

  for (i=0; i<NUM_FONTS; i++) {
    if (fontEntries[i].info != NULL)
      XFreeFont(display, fontEntries[i].info);
    if (fontEntries[i].writeGC != 0)
      XFreeGC(display, fontEntries[i].writeGC);
    if (fontEntries[i].eraseGC != 0)
      XFreeGC(display, fontEntries[i].eraseGC);
  }
  menuFont = NULL;
  picFont  = NULL;
  textFont = NULL;
}





/*----------------------------------------------------------------------------
  wind_GetScreenOffsets:
----------------------------------------------------------------------------*/
void
wind_GetScreenOffsets(n, xofs, yofs)
int n, *xofs, *yofs;
{
  *xofs = fontEntries[n].scrXofs;
  *yofs = fontEntries[n].scrYofs;
}




/*----------------------------------------------------------------------------
  wind_GetTexOffsets:
----------------------------------------------------------------------------*/
void
wind_GetTexOffsets(n, xofs, yofs)
int n;
double *xofs, *yofs;
{
  *xofs = fontEntries[n].texXofs;
  *yofs = fontEntries[n].texYofs;
}




/*----------------------------------------------------------------------------
  wind_HandleExposure:
----------------------------------------------------------------------------*/
void
wind_HandleExposure(event)
XEvent event;
{
  if (event.xexpose.count == 0) {
    if (event.xexpose.window == picWind) {
      XClearWindow(display, picWind);
      objs_DrawAll();
    }
    else wind_PaintMenuWind(event.xexpose.window);
  }
}





/*----------------------------------------------------------------------------
  wind_Initialize:
    Open the display & fonts, create windows & cursor.
----------------------------------------------------------------------------*/
void
wind_Initialize()
{
  wind_OpenDisplay();
  wind_DefinePixels();
  wind_OpenFonts();
  wind_MakeSmallCrossCursor();
  wind_MakeGraphics();
  wind_MakePicture();
  wind_MakeText();
  wind_MakeMenus();

  XMapSubwindows(display, graphWind);
}



/*----------------------------------------------------------------------------
  wind_LookupFontFromTex:
----------------------------------------------------------------------------*/
XFontStruct *
wind_LookupFontFromTex(name)
char *name;
{
  int i;

  for (i=0; i<NUM_FONTS; i++)
    if (str_IsEqual(name, fontEntries[i].texName))
      return fontEntries[i].info;

  return NULL;
}




/*----------------------------------------------------------------------------
  wind_LookupFontNum:
----------------------------------------------------------------------------*/
int
wind_LookupFontNum(font)
XFontStruct *font;
{
  int i;

  if (font == NULL)
    return NONE;

  for (i=0; i<NUM_FONTS; i++)
    if (font == fontEntries[i].info)
      return i;

  return NONE;
}




/*----------------------------------------------------------------------------
  wind_LookupMenuNum:
    Given a menu window, returns its number.
----------------------------------------------------------------------------*/
int
wind_LookupMenuNum(w)
Window w;
{
  int i, j;

  for (i=0; i<MENU_ROWS; i++)
    for (j=0; j<MENU_COLS; j++)
      if (menuWind[i][j] == w)
        return (i * MENU_COLS + j);
  return NONE;
}



/*----------------------------------------------------------------------------
  wind_LookupMenuWindow:
    Given a menu number, returns its window.
----------------------------------------------------------------------------*/
Window
wind_LookupMenuWindow(menuNum)
int menuNum;
{
  int i = menuNum / MENU_COLS;
  int j = menuNum % MENU_COLS;

  if (menuNum >= 0 && menuNum < NUM_MENUS) return menuWind[i][j];
  return NONE;
}




/*----------------------------------------------------------------------------
  wind_LookupTexFromFont:
----------------------------------------------------------------------------*/
char *
wind_LookupTexFromFont(font)
XFontStruct *font;
{
  int i;

  for (i=0; i<NUM_FONTS; i++)
    if (font == fontEntries[i].info)
      return fontEntries[i].texName;

  return NULL;
}




/* --------------------------------------------------------------------------
   wind_LookupTextGC:
---------------------------------------------------------------------------*/
GC
wind_LookupTextGC(font)  
XFontStruct *font;
{
  int i;

  for (i=0; i<NUM_FONTS; i++)
    if (font == fontEntries[i].info)
      return fontEntries[i].writeGC;

  return 0;
}




/* --------------------------------------------------------------------------
   wind_LookupTextEraseGC:
---------------------------------------------------------------------------*/
GC
wind_LookupTextEraseGC(font)  
XFontStruct *font;
{
  int i;

  for (i=0; i<NUM_FONTS; i++)
    if (font == fontEntries[i].info)
      return fontEntries[i].eraseGC;

  return 0;
}




/*----------------------------------------------------------------------------
  wind_MakeGraphics:
    Creates and maps the graphics window.
----------------------------------------------------------------------------*/
void
wind_MakeGraphics()
{
  int w = 16;
  int h = 16;
  int d = DefaultDepth(display, screen);
  int rw, rh;

  Pixmap backgroundPixmap;                    /* background shading pattern */
  XSetWindowAttributes winAttr;
  int attrMask = 0;

  /* background bitmap */
  static char background[] = {
   0x88, 0x88, 0x00, 0x00, 0x22, 0x22, 0x00, 0x00,
   0x88, 0x88, 0x00, 0x00, 0x22, 0x22, 0x00, 0x00,
   0x88, 0x88, 0x00, 0x00, 0x22, 0x22, 0x00, 0x00,
   0x88, 0x88, 0x00, 0x00, 0x22, 0x22, 0x00, 0x00};

  if (DefaultDepth(display, screen) > 1) {
    Pixmap srcP = XCreateBitmapFromData(display, rootWind, background, w, h);
    Pixmap destP = XCreatePixmap(display, rootWind, w, h, d);
    XCopyPlane(display, srcP, destP, DefaultGC(display, screen),
		0, 0, w, h, 0, 0, 1);
    winAttr.background_pixmap = destP;
  }
  else {
    winAttr.background_pixmap
	= XCreateBitmapFromData(display, rootWind, background, w, h);
  }
  attrMask |= CWBackPixmap;
  winAttr.border_pixel = brdr_pixel;
  attrMask |= CWBorderPixel;
  winAttr.event_mask = ExposureMask;
  attrMask |= CWEventMask;
  winAttr.cursor = smallCrossCursor;
  attrMask |= CWCursor;
  graphWind = XCreateWindow(display, rootWind,
			0, 0, GR_WIDTH, GR_HEIGHT, GR_BORDER, 0,
			InputOutput, CopyFromParent, attrMask, &winAttr);
  XStoreName(display, graphWind, programName);
  XMapWindow(display, graphWind);
}





/*----------------------------------------------------------------------------
  wind_MakeMenus:
    Opens NUM_MENUS menu windows.
----------------------------------------------------------------------------*/
void
wind_MakeMenus()
{
  Window w;
  int i, j;
  XSetWindowAttributes winAttr;
  int attrMask = 0;

  /* Define the menu windows (size, location, etc. */
  for (i=0; i<MENU_ROWS; i++) {
    for (j=0; j<MENU_COLS; j++) {
      winAttr.background_pixel = back_pixel;
      attrMask |= CWBackPixel;
      winAttr.border_pixel = brdr_pixel;
      attrMask |= CWBorderPixel;
      winAttr.win_gravity = SouthWestGravity;
      attrMask |= CWWinGravity;
      winAttr.event_mask = ButtonPressMask|EnterWindowMask
				|LeaveWindowMask|ExposureMask;
      attrMask |= CWEventMask;
      winAttr.cursor = smallCrossCursor;
      attrMask |= CWCursor;
      w = menuWind[i][j] = XCreateWindow(display, graphWind,
			j*(MENU_WIDTH+MENU_BORDER), i*(MENU_HEIGHT+MENU_BORDER),
			MENU_WIDTH, MENU_HEIGHT,
			1, 0, InputOutput, CopyFromParent, attrMask, &winAttr);

      XSync(display, 0);
    }
  }

  gc_MakeSelect();

  /* Enter text in the menu windows */
  for (i=0; i<NUM_MENUS; i++) {
    wind_UnselectMenu(i);
  }
}





/*----------------------------------------------------------------------------
  wind_MakePicture:
    Creates and maps the picture window.
----------------------------------------------------------------------------*/
void
wind_MakePicture()
{
  int picY;
  XSetWindowAttributes winAttr;
  int attrMask = 0;

  picHeight = ROUND(500 / scale);                                /* 500 pts */
  picWidth  = ROUND(400 / scale);                                /* 400 pts */
  picY = TEXT_Y - picHeight - PIC_BORDER * 2;

  winAttr.background_pixel = back_pixel;
  attrMask |= CWBackPixel;
  winAttr.border_pixel = brdr_pixel;
  attrMask |= CWBorderPixel;
  winAttr.bit_gravity = SouthWestGravity;
  attrMask |= CWBitGravity;
  winAttr.win_gravity = SouthWestGravity;
  attrMask |= CWWinGravity;
  winAttr.event_mask = ButtonPressMask|ButtonReleaseMask|KeyPressMask|
         Button1MotionMask|EnterWindowMask|LeaveWindowMask|ExposureMask;
  attrMask |= CWEventMask;
  winAttr.cursor = smallCrossCursor;
  attrMask |= CWCursor;
  picWind = XCreateWindow(display, graphWind,
			PIC_X, picY, picWidth, picHeight, PIC_BORDER, 0,
			InputOutput, CopyFromParent, attrMask, &winAttr);
  XSync(display, 0);

  gc_MakeTiled();
  gc_MakeGrid();
  gc_MakeThick();
  gc_MakeThickErase();
  gc_MakeThin();
  gc_MakeThinErase();
  gc_MakeFilled();
  gc_MakeFilledErase();
  gc_MakeXorDotted();
  gc_MakeXorDashed();

  obj_SetGridTile();
}





/*----------------------------------------------------------------------------
  wind_MakeSmallCrossCursor:
----------------------------------------------------------------------------*/
void
wind_MakeSmallCrossCursor()
{
  Font fid = XLoadFont(display, cursorFontName);

  smallCrossCursor = XCreateFontCursor(display, XC_tcross);
}



/*----------------------------------------------------------------------------
  wind_MakeText:
    Creates and maps the text window.
----------------------------------------------------------------------------*/
void
wind_MakeText()
{
  XSetWindowAttributes winAttr;
  int attrMask = 0;

  winAttr.background_pixel = back_pixel;
  attrMask |= CWBackPixel;
  winAttr.border_pixel = brdr_pixel;
  attrMask |= CWBorderPixel;
  winAttr.win_gravity = SouthWestGravity;
  attrMask |= CWWinGravity;
  winAttr.event_mask = KeyPressMask|EnterWindowMask
			|LeaveWindowMask|ExposureMask;
  attrMask |= CWEventMask;
  winAttr.cursor = smallCrossCursor;
  attrMask |= CWCursor;
  textWind = XCreateWindow(display, graphWind,
			TEXT_X, TEXT_Y, TEXT_WIDTH, TEXT_HEIGHT, TEXT_BORDER, 0,
			InputOutput, CopyFromParent, attrMask, &winAttr);
  XSync(display, 0);

  wind_ReportPictureSize();
}





/*----------------------------------------------------------------------------
  wind_Message1:
    Display a one-string error message in the text window.
----------------------------------------------------------------------------*/
void
wind_Message1(str)
char str[];
{
  XClearWindow(display, textWind);
  XDrawString(display, textWind, wind_LookupTextGC(textFont),
						10, 15, str, str_Length(str));
}




/*----------------------------------------------------------------------------
  wind_Message2:
    Display a two-string error message in the text window.
----------------------------------------------------------------------------*/
void
wind_Message2(str1, str2)
char str1[], str2[];
{
  int x = 10;
  int y = 15;

  XClearWindow(display, textWind);
  XDrawString(display, textWind, wind_LookupTextGC(textFont),
						x, y, str1, str_Length(str1));
  x += XTextWidth(textFont, str1, str_Length(str1));
  XDrawString(display, textWind, wind_LookupTextGC(textFont),
						x, y, str2, str_Length(str2));
}




/*----------------------------------------------------------------------------
  wind_OpenDisplay:
    Opens the X display and assigns the error handler functions.
----------------------------------------------------------------------------*/
void
wind_OpenDisplay()
{
  if ((display = XOpenDisplay(displayName)) == NULL) {
    printf("can not open display \n");
    exit(1);
  }
  screen = DefaultScreen(display);
}



/*----------------------------------------------------------------------------
  wind_OpenFonts:
    Opens the X fonts.
----------------------------------------------------------------------------*/
void
wind_OpenFonts()
{
  int i;

  for (i=0; i<NUM_FONTS; i++) {
    fontEntries[i].info = XLoadQueryFont(display, fontEntries[i].xName);
    fontEntries[i].writeGC = gc_MakeTextWrite(fontEntries[i].info->fid);
    fontEntries[i].eraseGC = gc_MakeTextErase(fontEntries[i].info->fid);
  }

  if (fontEntries[0].info == NULL) {
    printf("error:  font \"%s\" not found \n", fontEntries[0].xName);
    exit(1);
  }

  menuFont = fontEntries[0].info;
  textFont = fontEntries[0].info;

  wind_SetPicFont(0);

}




/*----------------------------------------------------------------------------
  wind_PaintMenu:
    1.  Searches "menuEntries" to find an entry with a "num" field
        matching "n".
    2.  If the menu is "selected", shades the background,
        otherwise clears the background to white.
    3.  If it is the FONTS menu, displays "Font" in the current
        picture font, otherwise displays line1 and line2 from
        menuEntries using the menu font.
----------------------------------------------------------------------------*/
void
wind_PaintMenu(n, w)
int n;
Window w;
{
  MenuEntry e;
  int i;

  if (menuSelected[n]) {
    XFillRectangle(display, w, selectGC, 0, 0, MENU_WIDTH, MENU_HEIGHT);
  }
  else XClearWindow(display, w);

  for (i=0; i<NUM_ENTRIES; i++) {
    e = menuEntries[i];
    if (e.num == n) {
      w = wind_LookupMenuWindow(n);
      if (n == FONTS) {
        int x, y, fontNum;
        fontNum = wind_LookupFontNum(picFont);
        switch(fontNum) {
          case 0:  x = 13;  y = 26;  break;
          case 1:  x = 13;  y = 27;  break;
          case 2:  x =  9;  y = 28;  break;
          case 3:  x =  7;  y = 29;  break;
          default: x =  7;  y = 29;  break;
        }
	XDrawString(display, w, wind_LookupTextGC(picFont), x, y, "FONT", 4);
      }
      else if (e.lines == 1)
	XDrawImageString(display, w, wind_LookupTextGC(menuFont),
							4, 26, e.line1, 7);
      else if (e.lines == 2) {
	XDrawImageString(display, w, wind_LookupTextGC(menuFont),
							4, 21, e.line1, 7);
	XDrawImageString(display, w, wind_LookupTextGC(menuFont),
							4, 31, e.line2, 7);
      }  /* else */
    }  /* if */
  }  /* for */
}





/*----------------------------------------------------------------------------
  wind_PaintMenuNum:
----------------------------------------------------------------------------*/
void
wind_PaintMenuNum(n)
int n;
{
  Window w;

  if (VALID(w = wind_LookupMenuWindow(n)))
    wind_PaintMenu(n, w);
}





/*----------------------------------------------------------------------------
  wind_PaintMenuWind:
----------------------------------------------------------------------------*/
void
wind_PaintMenuWind(w)
Window w;
{
  int n;

  if (VALID(n = wind_LookupMenuNum(w)))
    wind_PaintMenu(n, w);
}





/*----------------------------------------------------------------------------
            ***  TEMPORARY  ***
----------------------------------------------------------------------------*/
void
wind_PrintXFontStruct(font, name)
XFontStruct *font;
char name[];
{
  int i, n;
  char lc[2];
  char uc[2];

  lc[0] = 'a';
  lc[1] = '\0';
  uc[0] = 'A';
  uc[1] = '\0';
  printf("  font name: %s  \n", name);
  printf("     ascent: %4d \n", font->ascent);
  printf("    descent: %4d \n", font->descent);
  printf("      widths: \n");
  for (i=0; i<26; i++) {
    n = XTextWidth(font, lc, str_Length(lc));
    printf("                %s %d     ", lc, n);
    n = XTextWidth(font, lc, str_Length(uc));
    printf("%s %d \n", uc, n);
    lc[0] += 1;
    uc[0] += 1;
  }
  printf("\n");
}




/*----------------------------------------------------------------------------
  wind_PrintWindow:
    Print the name of a window.
----------------------------------------------------------------------------*/
void
wind_PrintWindow(w)
Window w;
{
  int n;

  if      (w == rootWind)   printf("rootWind");
  else if (w == textWind)   printf("textWind");
  else if (w == picWind)    printf("picWind");
  else if (w == graphWind)  printf("graphWind");
  else if (VALID(n = wind_LookupMenuNum(w)))
    printf("menu window %d", n);
  else
    printf("unknown window");
}




/*----------------------------------------------------------------------------
  wind_ReportPictureSize:
----------------------------------------------------------------------------*/
void
wind_ReportPictureSize()
{
  char message[80];
  int len;
  double w = ROUND2(scale * unitLength * picWidth);
  double h = ROUND2(scale * unitLength * picHeight);

  sprintf(message, "picture size:  %.2fW x %.2fH", w, h);
  len = str_Length(message);
  XClearWindow(display, textWind);
  XDrawString(display, textWind, wind_LookupTextGC(textFont),
							10, 10, message, len);
}




/*----------------------------------------------------------------------------
  wind_ResizePicture:
----------------------------------------------------------------------------*/
void
wind_ResizePicture(w, h)
int w, h;
{
  int picY;
  int textY;
  XWindowAttributes winAttr;
  
  XGetWindowAttributes(display, graphWind, &winAttr);
  textY = winAttr.height - TEXT_HEIGHT - TEXT_BORDER * 2;

  picWidth = MAX(MIN_WIND, w);
  picHeight = MAX(MIN_WIND, h);
  picY = textY - picHeight - PIC_BORDER * 2;

  XMoveResizeWindow(display, picWind, PIC_X, picY, picWidth, picHeight);
  obj_SetGridTile();
  wind_ReportPictureSize();
}





/*----------------------------------------------------------------------------
  wind_SelectMenu:
----------------------------------------------------------------------------*/
void
wind_SelectMenu(n)
int n;
{
  menuSelected[n] = TRUE;
  wind_PaintMenuNum(n);
}





/*----------------------------------------------------------------------------
  wind_SetPicFont:
----------------------------------------------------------------------------*/
void
wind_SetPicFont(n)
int n;
{

  if (fontEntries[n].info == NULL) {
    wind_ErrorMsg1("error - no font");
    return;
  }

  currFontNum = n;
  picFont = fontEntries[currFontNum].info;
  scrXofs = fontEntries[currFontNum].scrXofs;
  scrYofs = fontEntries[currFontNum].scrYofs;
  texXofs = fontEntries[currFontNum].texXofs;
  texYofs = fontEntries[currFontNum].texYofs;
}




/*----------------------------------------------------------------------------
  wind_UnselectMenu:
----------------------------------------------------------------------------*/
void
wind_UnselectMenu(n)
int n;
{
  /* Wait until one command is done executing before starting another. */
  XSync(display, 0);

  menuSelected[n] = FALSE;
  wind_PaintMenuNum(n);

}

